from __future__ import annotations

from collections.abc import Sequence
from contextlib import suppress
from typing import (
    Union,
    overload,
)

import pint

UnitsLike = Union[None, str, pint.Unit, Sequence[str | pint.Unit]]
UnitsInfo = Union[None, pint.Unit, tuple[pint.Unit, ...]]


__all__ = (
    'UnitsInfo',
    'UnitsLike',
    'get_units_from_name',
)


def get_unit_from_name(unit: str | pint.Unit | None) -> pint.Unit:
    if isinstance(unit, pint.Unit):
        return unit
    if unit is None:
        return pint.get_application_registry().pixel
    if isinstance(unit, str):
        with suppress(pint.errors.UndefinedUnitError):
            return pint.get_application_registry().parse_expression(unit).units

    raise ValueError(f'Could not find unit {unit}')


@overload
def get_units_from_name(units: None) -> None: ...


@overload
def get_units_from_name(units: str | pint.Unit | None) -> pint.Unit: ...


@overload
def get_units_from_name(
    units: Sequence[str | pint.Unit | None],
) -> tuple[pint.Unit, ...]: ...


def get_units_from_name(units: UnitsLike) -> UnitsInfo:
    """Convert a string or sequence of strings to pint units."""
    if isinstance(units, str) or units is None:
        return get_unit_from_name(units)
    if isinstance(units, Sequence):
        return tuple(get_unit_from_name(u) for u in units)
    raise ValueError(f'Could not find unit {units}')



import os
from pathlib import Path
from typing import Any

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from napari.settings._appearance import AppearanceSettings
from napari.settings._application import ApplicationSettings
from napari.settings._base import (
    _NOT_SET,
    EventedConfigFileSettings,
    _remove_empty_dicts,
)
from napari.settings._experimental import ExperimentalSettings
from napari.settings._fields import Version
from napari.settings._plugins import PluginsSettings
from napari.settings._shortcuts import ShortcutsSettings
from napari.utils._base import _DEFAULT_CONFIG_PATH
from napari.utils.translations import trans

_CFG_PATH = os.getenv('NAPARI_CONFIG', _DEFAULT_CONFIG_PATH)

CURRENT_SCHEMA_VERSION = Version(0, 9, 0)


class NapariSettings(EventedConfigFileSettings):
    """Schema for napari settings."""

    # 1. If you want to *change* the default value of a current option, you need to
    #    do a MINOR update in config version, e.g. from 3.0.0 to 3.1.0
    # 2. If you want to *remove* options that are no longer needed in the codebase,
    #    or if you want to *rename* options, then you need to do a MAJOR update in
    #    version, e.g. from 3.0.0 to 4.0.0
    # 3. You don't need to touch this value if you're just adding a new option
    schema_version: Version = Field(
        CURRENT_SCHEMA_VERSION,
        description=trans._('Napari settings schema version.'),
    )

    application: ApplicationSettings = Field(
        default_factory=ApplicationSettings,
        title=trans._('Application'),
        description=trans._('Main application settings.'),
    )
    appearance: AppearanceSettings = Field(
        default_factory=AppearanceSettings,
        title=trans._('Appearance'),
        description=trans._('User interface appearance settings.'),
        frozen=True,
    )
    plugins: PluginsSettings = Field(
        default_factory=PluginsSettings,
        title=trans._('Plugins'),
        description=trans._('Plugins settings.'),
        frozen=True,
    )
    shortcuts: ShortcutsSettings = Field(
        default_factory=ShortcutsSettings,
        title=trans._('Shortcuts'),
        description=trans._('Shortcut settings.'),
        frozen=True,
    )
    experimental: ExperimentalSettings = Field(
        default_factory=ExperimentalSettings,
        title=trans._('Experimental'),
        description=trans._('Experimental settings.'),
        frozen=True,
    )

    # private attributes and ClassVars will not appear in the schema
    config_path: Path | None = Field(
        Path(_CFG_PATH) if _CFG_PATH else None, exclude=True
    )

    model_config = SettingsConfigDict(
        env_prefix='napari_',
        nested_model_default_partial_update=True,
        env_nested_delimiter='_',
        env_nested_max_split=1,
        use_enum_values=False,
        extra='ignore',
        populate_by_name=True,
    )

    def __init__(self, config_path=_NOT_SET, **values: Any) -> None:
        super().__init__(config_path, **values)
        self._maybe_migrate()

    def _save_dict(self, **kwargs):
        # we always want schema_version written to the settings.yaml
        # TODO: is there a better way to always include schema version?
        return {
            'schema_version': self.schema_version,
            **super()._save_dict(**kwargs),
        }

    def __str__(self):
        out = 'NapariSettings (defaults excluded)\n' + 34 * '-' + '\n'
        data = self.model_dump(exclude_defaults=True)
        out += self._yaml_dump(_remove_empty_dicts(data))
        return out

    def __repr__(self):
        return str(self)

    def _maybe_migrate(self):
        if self.schema_version < CURRENT_SCHEMA_VERSION:
            from napari.settings._migrations import do_migrations

            do_migrations(self)


if __name__ == '__main__':
    import json
    import sys

    if len(sys.argv) > 2:
        dest = Path(sys.argv[2]).expanduser().absolute()
    else:
        dest = Path(__file__).parent / 'napari.schema.json'
    dest.write_text(json.dumps(NapariSettings.model_json_schema()))



"""Qt-based font rendering for vispy text visuals.

This module provides Qt-based FontManager and TextureFont implementations
that satisfy the vispy API but use Qt font machinery instead of FreeType.
This allows for uniform fonts across the entire napari application.
"""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

import numpy as np
from qtpy.QtCore import QPointF, Qt
from qtpy.QtGui import QFont, QFontMetricsF, QImage, QPainter
from vispy.gloo import TextureAtlas
from vispy.io import load_spatial_filters
from vispy.visuals.text.text import SDFRendererCPU

if TYPE_CHECKING:
    from typing import Any


def _load_glyph_qt(
    qfont: QFont,
    metrics: QFontMetricsF,
    char: str,
    glyphs_dict: dict[str, dict],
) -> None:
    """Load glyph from Qt font into dict.

    This is a Qt-based replacement for vispy's _load_glyph function
    that uses Qt's font rendering instead of FreeType.

    Parameters
    ----------
    qfont : QFont
        Qt font object to use for rendering.
    metrics : QFontMetricsF
        Qt font metrics object.
    char : str
        A single character to be represented.
    glyphs_dict : dict
        Dictionary to store glyph information.
    """

    # Calculate character dimensions and metrics
    bounding_rect = metrics.boundingRect(char)
    advance = metrics.horizontalAdvance(char)

    # Add padding to ensure we capture the full glyph
    padding = 4
    width = int(np.ceil(bounding_rect.width())) + 2 * padding
    height = int(np.ceil(bounding_rect.height())) + 2 * padding

    # Handle empty glyphs (e.g., space)
    if width <= 2 * padding or height <= 2 * padding:
        width = max(width, 2 * padding + 1)
        height = max(height, 2 * padding + 1)
        bitmap = np.zeros((height, width), dtype=np.uint8)
        left = 0
        top = int(metrics.ascent())
    else:
        # Create QImage to render the glyph
        image = QImage(width, height, QImage.Format.Format_Grayscale8)
        image.fill(Qt.GlobalColor.black)

        # Render the character
        painter = QPainter(image)
        painter.setFont(qfont)
        painter.setPen(Qt.GlobalColor.white)

        # Position the character in the image
        # Qt's drawText(x, y, text) places the baseline at y
        # bounding_rect.top() is negative for characters above baseline
        # bounding_rect.bottom() is positive for the bottom edge
        # We need to position so the bounding box is centered with padding
        x_pos = padding - bounding_rect.left()
        # For y: baseline should be at (padding - bounding_rect.top())
        # This puts the top of the bounding box at 'padding' pixels from top
        y_pos = padding - bounding_rect.top()
        painter.drawText(QPointF(x_pos, y_pos), char)
        painter.end()

        # Convert QImage to numpy array
        # QImage has stride/bytesPerLine that may not equal width due to alignment
        ptr = image.constBits()
        if ptr is None:
            raise RuntimeError(
                f'Failed to get image bits for character {char}'
            )
        if hasattr(ptr, 'setsize'):
            ptr.setsize(image.sizeInBytes())

        bytes_per_line = image.bytesPerLine()

        # Create array with proper stride, then extract only the actual image data
        # Type ignore: voidptr is a valid buffer but not recognized by mypy
        full_array = np.frombuffer(ptr, dtype=np.uint8).reshape(  # type: ignore[call-overload]
            (height, bytes_per_line)
        )
        bitmap = full_array[:, :width].copy()

        # Calculate glyph offset (left, top) in FreeType convention
        # left: horizontal offset from pen origin to left edge of bitmap
        # top: vertical offset from baseline to top edge of bitmap (positive = above baseline)
        # Our bitmap includes padding, so we need to account for that
        # The left edge of bitmap is at (bounding_rect.left() - padding)
        # The top edge of bitmap is at (-bounding_rect.top() - padding) from baseline
        left = int(bounding_rect.left() - padding)
        top = int(-bounding_rect.top() + padding)

    # Create glyph dict in vispy format
    kerning_dict: dict[str, float] = {}
    glyph: dict[str, Any] = {
        'char': char,
        'offset': (left, top),
        'bitmap': bitmap,
        'advance': advance,
        'kerning': kerning_dict,
    }
    glyphs_dict[char] = glyph

    # Generate kerning information for all existing glyphs
    # Note: Qt doesn't provide direct kerning access like FreeType,
    # so we approximate it using the difference between advances
    for other_char, other_glyph in glyphs_dict.items():
        if other_char == char:
            continue
        # Qt kerning approximation
        pair_width = metrics.horizontalAdvance(other_char + char)
        expected_width = other_glyph['advance'] + advance
        kerning = pair_width - expected_width
        glyph['kerning'][other_char] = kerning

        # Reverse direction
        pair_width_rev = metrics.horizontalAdvance(char + other_char)
        expected_width_rev = advance + other_glyph['advance']
        kerning_rev = pair_width_rev - expected_width_rev
        other_glyph['kerning'][char] = kerning_rev


class QtTextureFont:
    """Gather a set of glyphs relative to a given font using Qt rendering.

    This is a Qt-based replacement for vispy's TextureFont that uses
    Qt's font rendering machinery instead of FreeType.

    Parameters
    ----------
    font : dict
        Dict with entries "face", "size", "bold", "italic".
    renderer : instance of SDFRenderer
        SDF renderer to use.
    """

    def __init__(self, font: dict[str, Any], renderer: SDFRendererCPU) -> None:
        self._atlas = TextureAtlas(dtype=np.uint8)
        self._atlas.wrapping = 'clamp_to_edge'
        self._kernel, _ = load_spatial_filters()
        self._renderer = renderer
        self._font = deepcopy(font)
        self._font['size'] = 256  # use high resolution point size for SDF
        self._lowres_size = 64  # end at this point size for storage
        assert (self._font['size'] % self._lowres_size) == 0
        # spread/border at the high-res for SDF calculation
        self._spread = 32
        assert self._spread % self.ratio == 0
        self._glyphs: dict[str, dict[str, Any]] = {}

        # Create and cache Qt font and metrics objects
        self._qfont = QFont(self._font['face'], self._font['size'])
        self._qfont.setBold(self._font.get('bold', False))
        self._qfont.setItalic(self._font.get('italic', False))
        self._metrics = QFontMetricsF(self._qfont)

    @property
    def ratio(self) -> int:
        """Ratio of the initial high-res to final stored low-res glyph."""
        return self._font['size'] // self._lowres_size

    @property
    def slop(self) -> int:
        """Extra space along each glyph edge due to SDF borders."""
        return self._spread // self.ratio

    def __getitem__(self, char: str) -> dict[str, Any]:
        if not (isinstance(char, str) and len(char) == 1):
            raise TypeError('index must be a 1-character string')
        if char not in self._glyphs:
            self._load_char(char)
        return self._glyphs[char]

    def _load_char(self, char: str) -> None:
        """Build and store a glyph corresponding to an individual character.

        Parameters
        ----------
        char : str
            A single character to be represented.
        """
        assert isinstance(char, str)
        assert len(char) == 1
        assert char not in self._glyphs

        # Load new glyph data using Qt
        _load_glyph_qt(self._qfont, self._metrics, char, self._glyphs)

        # Put new glyph into the texture
        glyph = self._glyphs[char]
        bitmap = glyph['bitmap']

        # Convert to padded array
        data = np.zeros(
            (
                bitmap.shape[0] + 2 * self._spread,
                bitmap.shape[1] + 2 * self._spread,
            ),
            np.uint8,
        )
        data[self._spread : -self._spread, self._spread : -self._spread] = (
            bitmap
        )

        # Store, while scaling down to proper size
        height = data.shape[0] // self.ratio
        width = data.shape[1] // self.ratio
        region = self._atlas.get_free_region(width + 2, height + 2)
        if region is None:
            raise RuntimeError('Cannot store glyph')
        x, y, w, h = region
        x, y, w, h = x + 1, y + 1, w - 2, h - 2

        self._renderer.render_to_texture(data, self._atlas, (x, y), (w, h))
        u0 = x / float(self._atlas.shape[1])
        v0 = y / float(self._atlas.shape[0])
        u1 = (x + w) / float(self._atlas.shape[1])
        v1 = (y + h) / float(self._atlas.shape[0])
        texcoords = (u0, v0, u1, v1)
        glyph.update({'size': (w, h), 'texcoords': texcoords})


class QtFontManager:
    """Helper to create QtTextureFont instances and reuse them when possible.

    This is a Qt-based replacement for vispy's FontManager that uses
    Qt's font rendering machinery.

    Parameters
    ----------
    method : str, optional
        Rendering method ('cpu' or 'gpu'). Default is 'cpu'.
        Currently only 'cpu' is supported for Qt-based rendering.
    """

    def __init__(self, method: str = 'cpu') -> None:
        self._fonts: dict[str, QtTextureFont] = {}
        if not isinstance(method, str) or method not in ('cpu', 'gpu'):
            raise ValueError(
                f'method must be "cpu" or "gpu", got {method} ({type(method)})'
            )
        if method == 'cpu':
            self._renderer = SDFRendererCPU()
        else:  # method == 'gpu':
            from vispy.visuals.text.text import SDFRendererGPU

            self._renderer = SDFRendererGPU()

    def get_font(
        self, face: str, bold: bool = False, italic: bool = False
    ) -> QtTextureFont:
        """Get a font described by face and size.

        Parameters
        ----------
        face : str
            Font face name.
        bold : bool, optional
            Whether to use bold weight.
        italic : bool, optional
            Whether to use italic style.

        Returns
        -------
        QtTextureFont
            The texture font instance.
        """
        key = f'{face}-{bold}-{italic}'
        if key not in self._fonts:
            font = {'face': face, 'bold': bold, 'italic': italic}
            self._fonts[key] = QtTextureFont(font, self._renderer)
        return self._fonts[key]



from __future__ import annotations

import logging
import weakref
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Protocol

    from napari.utils.events.event import EmitterGroup

    class Emitter(Protocol):
        def connect(self, callback: Callable): ...

        def disconnect(self, callback: Callable): ...


_logger = logging.getLogger(__name__)


def disconnect_events(emitter: EmitterGroup, listener: object) -> None:
    """Disconnect all events between an emitter group and a listener.

    Parameters
    ----------
    emitter : napari.utils.events.event.EmitterGroup
        Emitter group.
    listener : Object
        Any object that has been connected to.
    """
    for em in emitter.emitters.values():
        em.disconnect(listener)


def connect_setattr(
    emitter: Emitter,
    obj,
    attr: str,
    convert_fun: Callable[[Any], Any] | None = None,
) -> None:
    ref = weakref.ref(obj)
    if convert_fun:
        # Handle passed `convert_func` function to map emitted values to valid
        # values accepted for the receiver object attribute.
        # A `convert_func` is needed to, for example, map `Qt.CheckState`
        # values to boolean ones when a `QCheckBox` value change is connected
        # to a layer attribute.
        # See napari/napari#8154
        def _cb(*value):
            if (ob := ref()) is None:
                emitter.disconnect(_cb)
                return

            value = tuple(convert_fun(x) for x in value)
            setattr(ob, attr, value[0] if len(value) == 1 else value)
    else:

        def _cb(*value):
            if (ob := ref()) is None:
                emitter.disconnect(_cb)
                return

            setattr(ob, attr, value[0] if len(value) == 1 else value)

    emitter.connect(_cb)
    # There are scenarios where emitter is deleted before obj.
    # Also there is no option to create weakref to QT Signal
    # but even if keep reference to base object and signal name it is possible to meet
    # problem with C++ "wrapped C/C++ object has been deleted"

    # In all of these 3 functions, this should be uncommented instead of using
    # the if clause in _cb but that causes a segmentation fault in tests
    # weakref.finalize(obj, emitter.disconnect, _cb)


def connect_no_arg(emitter: Emitter, obj, attr: str):
    ref = weakref.ref(obj)

    def _cb(*_value):
        if (ob := ref()) is None:
            emitter.disconnect(_cb)
            return
        getattr(ob, attr)()

    emitter.connect(_cb)
    # as in connect_setattr
    # weakref.finalize(obj, emitter.disconnect, _cb)


def connect_setattr_value(emitter: Emitter, obj, attr: str):
    """To get value from Event"""
    ref = weakref.ref(obj)

    def _cb(value):
        if (ob := ref()) is None:
            emitter.disconnect(_cb)
            return
        setattr(ob, attr, value.value)

    emitter.connect(_cb)
    # weakref.finalize(obj, emitter.disconnect, _cb)



import numpy as np

from napari.components import Dims
from napari.layers import Shapes
from napari.layers.shapes import _shapes_key_bindings as key_bindings


def test_lock_aspect_ratio():
    # Test a single four corner rectangle
    layer = Shapes(20 * np.random.random((1, 4, 2)))
    layer._moving_coordinates = (0, 0, 0)
    layer._is_moving = True
    # need to go through the generator
    _ = list(key_bindings.hold_to_lock_aspect_ratio(layer))


def test_lock_aspect_ratio_selected_box():
    # Test a single four corner rectangle
    layer = Shapes(20 * np.random.random((1, 4, 2)))
    # select a shape
    layer._selected_box = layer.interaction_box(0)
    layer._moving_coordinates = (0, 0, 0)
    layer._is_moving = True
    # need to go through the generator
    _ = list(key_bindings.hold_to_lock_aspect_ratio(layer))


def test_lock_aspect_ratio_selected_box_zeros():
    # Test a single four corner rectangle that has zero size
    layer = Shapes(20 * np.zeros((1, 4, 2)))
    # select a shape
    layer._selected_box = layer.interaction_box(0)
    layer._moving_coordinates = (0, 0, 0)
    layer._is_moving = True
    # need to go through the generator
    _ = list(key_bindings.hold_to_lock_aspect_ratio(layer))


def test_activate_modes():
    # Test a single four corner rectangle
    layer = Shapes(20 * np.random.random((1, 4, 2)))
    # need to go through the generator
    key_bindings.activate_add_rectangle_mode(layer)
    assert layer.mode == 'add_rectangle'
    key_bindings.activate_add_ellipse_mode(layer)
    assert layer.mode == 'add_ellipse'
    key_bindings.activate_add_line_mode(layer)
    assert layer.mode == 'add_line'
    key_bindings.activate_add_polyline_mode(layer)
    assert layer.mode == 'add_polyline'
    key_bindings.activate_add_path_mode(layer)
    assert layer.mode == 'add_path'
    key_bindings.activate_add_polygon_mode(layer)
    assert layer.mode == 'add_polygon'
    key_bindings.activate_direct_mode(layer)
    assert layer.mode == 'direct'
    key_bindings.activate_select_mode(layer)
    assert layer.mode == 'select'
    key_bindings.activate_shapes_pan_zoom_mode(layer)
    assert layer.mode == 'pan_zoom'
    key_bindings.activate_vertex_insert_mode(layer)
    assert layer.mode == 'vertex_insert'
    key_bindings.activate_vertex_remove_mode(layer)
    assert layer.mode == 'vertex_remove'


def test_copy_paste():
    # Test on three four corner rectangle
    layer = Shapes(20 * np.random.random((3, 4, 2)))
    layer.mode = 'direct'

    assert len(layer.data) == 3
    assert layer._clipboard == {}
    layer.selected_data = {0, 1}

    key_bindings.copy_selected_shapes(layer)
    assert len(layer.data) == 3
    assert len(layer._clipboard) > 0

    key_bindings.paste_shape(layer)
    assert len(layer.data) == 5
    assert len(layer._clipboard) > 0


def test_select_shapes_in_slice():
    # Test on three four corner rectangle
    data = np.array(
        [
            [[0, 0], [0, 10], [10, 10], [10, 0]],
            [[20, 20], [20, 30], [30, 30], [30, 20]],
            [[40, 40], [40, 50], [50, 50], [50, 40]],
        ]
    )
    layer = Shapes(data)
    layer.mode = 'direct'

    assert len(layer.data) == 3
    assert len(layer.selected_data) == 0

    key_bindings.select_shapes_in_slice(layer)
    assert len(layer.selected_data) == 3

    # Calling it again should deselect all
    key_bindings.select_shapes_in_slice(layer)
    assert len(layer.selected_data) == 0


def test_select_shapes_in_slice_3d():
    """Test select all shapes in a 3D slice."""
    data = [
        np.array([[0, 0, 0], [0, 10, 10]]),  # shape 0 on slice 0
        np.array([[1, 5, 5], [1, 15, 15]]),  # shape 1 on slice 1
        np.array([[0, 20, 20], [0, 30, 30]]),  # shape 2 on slice 0
    ]
    layer = Shapes(data, shape_type='line')
    layer.mode = 'select'

    # view slice 0
    layer._slice_dims(
        Dims(
            ndim=3,
            ndisplay=2,
            range=((0, 2, 1), (0, 40, 1), (0, 40, 1)),
            point=(0, 0, 0),
        )
    )

    # select all shapes on slice 0
    key_bindings.select_shapes_in_slice(layer)

    # Check that shapes 0 and 2 are selected
    assert layer.selected_data == {0, 2}

    # Check that an interaction box is created
    assert layer._selected_box is not None

    # Check that get_status doesn't crash
    status = layer.get_status(position=(0, 10, 10))
    assert isinstance(status, dict)


def test_delete():
    # Test on three four corner rectangle
    layer = Shapes(20 * np.random.random((3, 4, 2)))
    layer.mode = 'direct'

    assert len(layer.data) == 3
    layer.selected_data = {0, 1}

    key_bindings.delete_selected_shapes(layer)
    assert len(layer.data) == 1


def test_finish():
    # Test on three four corner rectangle
    layer = Shapes(20 * np.random.random((3, 4, 2)))
    key_bindings.finish_drawing_shape(layer)



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
from qtpy.QtGui import QFont, QFontMetricsF, QGuiApplication

from napari.layers import Points, Shapes
from napari.layers.utils.string_encoding import ConstantStringEncoding

if TYPE_CHECKING:
    from napari._vispy.visuals.text import Text


def update_text(
    *,
    node: Text,
    layer: Points | Shapes,
):
    """Update the vispy text node with a layer's text parameters.

    Parameters
    ----------
    node : vispy.scene.visuals.Text
        The text node to be updated.
    layer : Union[Points, Shapes]
        A layer with text.
    """

    ndisplay = layer._slice_input.ndisplay

    # Vispy always needs non-empty values and coordinates, so if a layer
    # effectively has no visible text then return single dummy data.
    # This also acts as a minor optimization.
    if _has_visible_text(layer):
        text_values = layer._view_text
        colors = layer._view_text_color
        coords, anchor_x, anchor_y = layer._view_text_coords
    else:
        text_values = np.array([''])
        colors = np.zeros((4,), np.float32)
        coords = np.zeros((1, ndisplay))
        anchor_x = 'center'
        anchor_y = 'center'

    # Vispy wants (x, y) positions instead of (row, column) coordinates.
    if ndisplay == 2:
        positions = np.flip(coords, axis=1)
    elif ndisplay == 3:
        raw_positions = np.flip(coords, axis=1)
        n_positions, position_dims = raw_positions.shape

        if position_dims < 3:
            padded_positions = np.zeros((n_positions, 3))
            padded_positions[:, 0:2] = raw_positions
            positions = padded_positions
        else:
            positions = raw_positions

    node.text = text_values
    node.pos = positions
    node.anchors = (anchor_x, anchor_y)

    text_manager = layer.text
    node.rotation = text_manager.rotation
    node.color = colors

    node.font_size = text_manager._get_scaled_size(layer.scale_factor)


def _has_visible_text(layer: Points | Shapes) -> bool:
    text = layer.text
    if not text.visible:
        return False
    if (
        isinstance(text.string, ConstantStringEncoding)
        and text.string.constant == ''
    ):
        return False
    return len(layer._indices_view) != 0


@lru_cache(maxsize=128)
def _get_qt_font_metrics(
    face: str, size: int, bold: bool = False, italic: bool = False
):
    """Get cached Qt font metrics for the given font properties.

    Parameters
    ----------
    face : str
        Font face name.
    size : int
        Font size in points.
    bold : bool, optional
        Whether the font is bold.
    italic : bool, optional
        Whether the font is italic.

    Returns
    -------
    QFontMetricsF
        Qt font metrics object.
    """
    qfont = QFont(face, size)
    qfont.setBold(bold)
    qfont.setItalic(italic)
    return QFontMetricsF(qfont)


def get_text_metrics(text: Text) -> QFontMetricsF:
    """Get qt font metrics from a text visual."""
    face = (
        text.face if hasattr(text, 'face') else QGuiApplication.font().family()
    )
    bold = text.bold if hasattr(text, 'bold') else False
    italic = text.italic if hasattr(text, 'italic') else False

    return _get_qt_font_metrics(face, int(text.font_size), bold, italic)


def get_text_width_height(text: Text) -> tuple[float, float]:
    """Get the width and height of a vispy text visual in screen pixels.

    If display scaling is not 1 (e.g. hidpi), this is already accounted for
    by vispy.
    """
    if isinstance(text.text, str):
        strings = [text.text]
    elif isinstance(text.text, list):
        strings = text.text
    else:
        raise TypeError('Text should either be a string or a list of strings')

    metrics = get_text_metrics(text)
    font_height = metrics.height()

    height = 0.0
    width = 0.0

    for string in strings:
        if string == '':
            continue

        for lineno, line in enumerate(string.split('\n')):
            width = max(width, metrics.horizontalAdvance(line))
            height_lines = font_height * (lineno + 1)
            height_line_spacing = font_height * (text.line_height - 1) * lineno
            height = max(height, height_lines + height_line_spacing)

    return width, height



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from vispy.scene.visuals import Rectangle
from vispy.visuals.transforms import MatrixTransform, STTransform

from napari._vispy.utils.gl import BLENDING_MODES
from napari.settings import get_settings
from napari.utils.color import ColorValue
from napari.utils.events import disconnect_events
from napari.utils.theme import get_theme

if TYPE_CHECKING:
    from vispy.scene import Node, ViewBox
    from vispy.visuals.text.text import FontManager

    from napari.components.overlays import CanvasOverlay, Overlay, SceneOverlay
    from napari.components.viewer_model import ViewerModel
    from napari.layers import Layer
    from napari.utils.events import Event


class VispyBaseOverlay:
    """
    Base overlay backend for vispy.

    Creates event connections between napari Overlay models and the
    vispy backend, translating them into rendering.
    """

    overlay: Overlay
    viewer: ViewerModel

    def __init__(
        self,
        *,
        overlay: Overlay,
        viewer: ViewerModel,
        node: Node,
        parent: ViewBox | None = None,
        font_manager: FontManager | None = None,
        font_family: str = 'OpenSans',
    ) -> None:
        super().__init__()
        self.overlay = overlay
        self.viewer = viewer
        self.font_manager = font_manager
        self.font_family = font_family

        self.node = node
        self.node.order = self.overlay.order

        self.overlay.events.visible.connect(self._on_visible_change)
        self.overlay.events.opacity.connect(self._on_opacity_change)
        self.overlay.events.blending.connect(self._on_blending_change)

        if parent is not None:
            self.node.parent = parent

    def _should_be_visible(self) -> bool:
        return self.overlay.visible

    def _on_visible_change(self) -> None:
        self.node.visible = self._should_be_visible()

    def _on_opacity_change(self) -> None:
        self.node.opacity = self.overlay.opacity

    def _on_blending_change(self) -> None:
        self.node.set_gl_state(**BLENDING_MODES[self.overlay.blending])
        self.node.update()

    def reset(self) -> None:
        self._on_visible_change()
        self._on_opacity_change()
        self._on_blending_change()

    def close(self) -> None:
        self.overlay.events.visible.disconnect(self._on_visible_change)
        self.overlay.events.opacity.disconnect(self._on_opacity_change)
        self.overlay.events.blending.disconnect(self._on_blending_change)
        disconnect_events(self.viewer.events, self)
        self.node.transforms = MatrixTransform()
        self.node.parent = None


class VispyCanvasOverlay(VispyBaseOverlay):
    """
    Vispy overlay backend for overlays that live in canvas space.

    NOTE: Subclasses must follow some rules:
    - ensure that when `_on_position_change` is called, the x_size and y_size
      attributes are already updated depending on the overlay size, to ensure
      proper tiling. Alternatively, override this method if the overlay is
      *not* supposed to be tiled
    - ensure that the napari Overlay model uses the `position` field correctly
      (must be a CanvasPosition enum if tileable, or anything else if "free")

    canvas_position_callback is set by the VispyCanvas object, and is responsible
    to update the position of all canvas overlays whenever necessary
    """

    overlay: CanvasOverlay

    def __init__(
        self, *, overlay, viewer, node, parent=None, **kwargs
    ) -> None:

        super().__init__(
            overlay=overlay, viewer=viewer, node=node, parent=parent, **kwargs
        )
        self.x_size = 0.0
        self.y_size = 0.0
        self.node.transform = STTransform()
        self.overlay.events.position.connect(self._on_position_change)
        self.overlay.events.box.connect(self._on_box_change)
        self.overlay.events.box_color.connect(self._on_box_change)
        get_settings().appearance.events.theme.connect(self._on_box_change)
        self.viewer.events.theme.connect(self._on_box_change)
        self.canvas_position_callback = lambda: None

        self.box = Rectangle(center=(0, 0), border_width=0)

    def _on_visible_change(self) -> None:
        super()._on_visible_change()
        self._on_box_change()

    def _on_box_change(self) -> None:
        if not self.overlay.box or not self._should_be_visible():
            self.box.parent = None
            return

        self.box.parent = self.node.parent

        # TODO: this should be related to tiling padding
        padding = 8
        self.box.width = self.x_size + padding
        self.box.height = self.y_size + padding
        self.box.center = self.x_size / 2, self.y_size / 2

        if self.overlay.box_color is None:
            bgcolor = self._get_canvas_bgcolor()
            # make the color a bit transparent
            bgcolor[-1] *= 0.8
        else:
            bgcolor = self.overlay.box_color

        self.box.color = bgcolor

        self.box.order = self.node.order - 1
        self.box.transform = self.node.transform

    def _get_canvas_bgcolor(self) -> ColorValue:
        if self.node.parent is not None and self.node.parent.canvas.bgcolor:
            return ColorValue(self.node.parent.canvas.bgcolor.rgba)

        return ColorValue(
            get_theme(get_settings().appearance.theme).canvas.as_rgb_tuple()
        )

    def _get_fgcolor(self) -> ColorValue:
        if not self.overlay.box or self.overlay.box_color is None:
            bgcolor = self._get_canvas_bgcolor()
        else:
            bgcolor = self.overlay.box_color
        return self._contrasting_color(bgcolor)

    def _contrasting_color(self, bgcolor: ColorValue) -> ColorValue:
        opposite = 1 - bgcolor
        # shift away from mid tones for better contrast
        opposite = 0.5 + (opposite - 0.5) * 1.2
        opposite = np.clip(opposite, 0, 1)
        # don't change alpha
        opposite[-1] = bgcolor[-1]
        return opposite

    def _on_blending_change(self) -> None:
        self.box.set_gl_state(**BLENDING_MODES[self.overlay.blending])
        super()._on_blending_change()

    def _on_position_change(self, event: Event | None = None) -> None:
        # NOTE: when subclasses call this method, they should first ensure sizes
        # (x_size, and y_size) are set correctly
        self._on_box_change()
        self.canvas_position_callback()

    def reset(self) -> None:
        super().reset()
        self._on_position_change()

    def close(self) -> None:
        super().close()
        self.box.parent = None
        self.canvas_position_callback = lambda: None


class VispySceneOverlay(VispyBaseOverlay):
    """
    Vispy overlay backend for overlays that live in scene (2D or 3D) space.
    """

    overlay: SceneOverlay

    def __init__(
        self, *, overlay, viewer, node, parent=None, **kwargs
    ) -> None:
        super().__init__(
            overlay=overlay, viewer=viewer, node=node, parent=parent, **kwargs
        )
        self.node.transform = MatrixTransform()


class LayerOverlayMixin:
    layer: Layer

    def __init__(
        self, *, overlay, layer: Layer, viewer, node, parent=None, **kwargs
    ) -> None:
        super().__init__(
            node=node,
            overlay=overlay,
            viewer=viewer,
            parent=parent,
            **kwargs,
        )
        self.layer = layer
        # need manual connection here because these overlays are not necessarily
        # always a child of the actual vispy node of the layer (eg, canvas overlays)
        self.layer.events.visible.connect(self._on_visible_change)

    def _should_be_visible(self) -> bool:
        return self.overlay.visible and self.layer.visible

    def close(self) -> None:
        disconnect_events(self.layer.events, self)
        super().close()


class ViewerOverlayMixin:
    pass



from __future__ import annotations

from itertools import cycle
from random import sample
from typing import TYPE_CHECKING, Any

import numpy as np
from vispy.app.timer import Timer
from vispy.visuals.transforms import NullTransform

from napari._vispy.overlays.base import ViewerOverlayMixin, VispyCanvasOverlay
from napari._vispy.visuals.welcome import Welcome
from napari.settings import get_settings

if TYPE_CHECKING:
    from vispy.scene import Node
    from vispy.util.event import Event
    from vispy.visuals.text.text import FontManager

    from napari.components import ViewerModel
    from napari.components.overlays import WelcomeOverlay


class VispyWelcomeOverlay(ViewerOverlayMixin, VispyCanvasOverlay):
    overlay: WelcomeOverlay

    def __init__(
        self,
        *,
        viewer: ViewerModel,
        overlay: WelcomeOverlay,
        font_manager: FontManager | None = None,
        font_family: str = 'OpenSans',
        parent: Node | None = None,
    ) -> None:
        super().__init__(
            node=Welcome(font_manager=font_manager, face=font_family),
            viewer=viewer,
            overlay=overlay,
            parent=parent,
            font_manager=font_manager,
            font_family=font_family,
        )
        self.viewer.events.theme.connect(self._on_theme_change)
        self.viewer.layers.events.inserted.connect(self._on_visible_change)
        self.viewer.layers.events.removed.connect(self._on_visible_change)

        get_settings().appearance.events.theme.connect(self._on_theme_change)

        self.overlay.events.version.connect(self._on_version_change)
        self.overlay.events.shortcuts.connect(self._on_shortcuts_change)
        self.overlay.events.tips.connect(self._on_tips_change)
        self.overlay.events.box.connect(self._on_theme_change)
        self.overlay.events.box_color.connect(self._on_theme_change)

        self.node.canvas.native.resized.connect(self._on_position_change)

        self.tips_iterator = cycle(["You're awesome!"])
        self.tip_timer = Timer(10, self.next_tip)
        self.next_tip()

        self.reset()

    def _should_be_visible(self) -> bool:
        return bool(super()._should_be_visible() and not self.viewer.layers)

    def _on_position_change(self, event: Any = None) -> None:
        if self.node.canvas is not None:
            x, y = np.array(self.node.canvas.size)
            self.node.set_scale_and_position(x, y)
            self.x_size = x
            self.y_size = y
        self._on_box_change()

    def _on_box_change(self) -> None:
        super()._on_box_change()
        # welcome uses some custom positioning with the transform, so copying it
        # over messes it up. We just set it to nothing.
        self.box.transform = NullTransform()
        # always opaque box color, so we hide what's behind
        bgcolor = self.box.color.rgba
        bgcolor[-1] = 1
        self.box.color = bgcolor

    def _on_theme_change(self) -> None:
        color = self._get_fgcolor()
        # dim a bit but keep opaque to avoid border artifacts
        color[:3] *= 0.7
        color[-1] = 1
        self.node.set_color(color)

    def _on_visible_change(self) -> None:
        show = self._should_be_visible()
        self.node.visible = show
        if show:
            self.tip_timer.start()
        else:
            try:
                self.tip_timer.stop()
            except RuntimeError as e:  # pragma: no cover
                if (
                    'wrapped C/C++ object of type' not in e.args[0]
                    and 'Internal C++ object' not in e.args[0]
                ):
                    # checking if the object is partially deleted. Otherwise
                    # reraise exception. For more details see:
                    # https://github.com/napari/napari/pull/5499
                    raise

    def _on_version_change(self) -> None:
        self.node.set_version(self.overlay.version)

    def _on_shortcuts_change(self) -> None:
        self.node.set_shortcuts(self.overlay.shortcuts)

    def _on_tips_change(self) -> None:
        if self.overlay.tips:
            self.tips_iterator = cycle(
                sample(self.overlay.tips, len(self.overlay.tips))
            )
        else:
            self.tips_iterator = cycle(["You're awesome!"])

    def next_tip(self, event: Event | None = None) -> None:
        self.node.set_tip(next(self.tips_iterator))

    def reset(self) -> None:
        super().reset()
        self._on_theme_change()
        self._on_version_change()
        self._on_shortcuts_change()
        self._on_tips_change()
        self.next_tip()

    def close(self) -> None:
        self.tip_timer.stop()
        super().close()



import numpy as np

from napari.components.overlays import (
    BoundingBoxOverlay,
    CanvasOverlay,
    ScaleBarOverlay,
)


def test_viewer_overlays(qt_viewer):
    viewer = qt_viewer.viewer
    canvas = qt_viewer.canvas

    for overlay in viewer._overlays.values():
        # vispy overlays only exist if they are visible at least once
        overlay.visible = True
        if isinstance(overlay, CanvasOverlay):
            assert all(
                visual.node in canvas.view.children
                for visual in canvas._overlay_to_visual[overlay]
            )
        else:
            assert all(
                visual.node in canvas.view.scene.children
                for visual in canvas._overlay_to_visual[overlay]
            )

    old_vispy_overlays = {
        k: list(v) for k, v in canvas._overlay_to_visual.items()
    }

    new_overlay = ScaleBarOverlay(visible=True)
    viewer._overlays['test'] = new_overlay

    assert new_overlay in canvas._overlay_to_visual
    new_overlay_node = canvas._overlay_to_visual[new_overlay][0].node
    assert new_overlay_node not in canvas.view.scene.children
    assert new_overlay_node in canvas.view.children

    # old visuals should still be there, as they are reused when possible
    for overlay, vispy_overlays in old_vispy_overlays.items():
        for vispy_overlay in vispy_overlays:
            if isinstance(overlay, CanvasOverlay):
                assert vispy_overlay.node in canvas.view.children
            else:
                assert vispy_overlay.node in canvas.view.scene.children

    viewer._overlays.pop('test')
    assert new_overlay not in canvas._overlay_to_visual
    assert new_overlay_node not in canvas.view.children

    viewer.welcome_screen.visible = False  # just for proper test cleanup


def test_layer_overlays(qt_viewer):
    viewer = qt_viewer.viewer
    canvas = qt_viewer.canvas

    view_children = len(canvas.view.children)
    scene_children = len(canvas.view.scene.children)

    assert not canvas._layer_overlay_to_visual

    layer = viewer.add_points()
    layer_node = canvas.layer_to_visual[layer].node

    for overlay in layer._overlays.values():
        # vispy overlays only exist if they are visible at least once
        overlay.visible = True
        if isinstance(overlay, CanvasOverlay):
            assert (
                canvas._layer_overlay_to_visual[layer][overlay].node
                in canvas.view.children
            )
        else:
            assert (
                canvas._layer_overlay_to_visual[layer][overlay].node
                in layer_node.children
            )

    old_vispy_overlays = {**canvas._layer_overlay_to_visual[layer]}

    new_overlay = BoundingBoxOverlay(visible=True)
    layer._overlays['test'] = new_overlay

    assert new_overlay in canvas._layer_overlay_to_visual[layer]
    new_overlay_node = canvas._layer_overlay_to_visual[layer][new_overlay].node
    assert new_overlay_node in layer_node.children
    assert new_overlay_node not in canvas.view.children

    # old visuals should still be there, as they are reused when possible
    for overlay, vispy_overlay in old_vispy_overlays.items():
        if isinstance(overlay, CanvasOverlay):
            assert vispy_overlay.node in canvas.view.children
        else:
            assert vispy_overlay.node in layer_node.children

    layer._overlays.pop('test')
    assert new_overlay not in canvas._layer_overlay_to_visual[layer]
    assert new_overlay_node not in canvas.view.children

    viewer.layers.pop()

    # should be back to the status quo
    assert not canvas._layer_overlay_to_visual
    assert len(canvas.view.children) == view_children
    assert len(canvas.view.scene.children) == scene_children


def test_grid_mode(qt_viewer):
    viewer = qt_viewer.viewer
    canvas = qt_viewer.canvas

    viewer.dims.ndisplay = 3
    viewer.add_image(np.ones((10, 10, 10)))

    angles = 10, 20, 30  # just some nonzero stuff
    zoom = 1
    viewer.camera.angles = angles
    viewer.camera.zoom = zoom

    canvas.on_draw(None)

    for camera in (canvas.camera, *canvas.grid_cameras):
        np.testing.assert_allclose(camera.angles, angles)
        assert camera.zoom == zoom

    # ensure that switching to grid maintains zoom and angles
    viewer.grid.enabled = True

    canvas.on_draw(None)

    for camera in (canvas.camera, *canvas.grid_cameras):
        np.testing.assert_allclose(camera.angles, angles)
        assert camera.zoom == zoom

    viewer.grid.enabled = False

    canvas.on_draw(None)

    for camera in (canvas.camera, *canvas.grid_cameras):
        np.testing.assert_allclose(camera.angles, angles)
        assert camera.zoom == zoom


def test_tiling_canvas_overlays(qt_viewer):
    viewer = qt_viewer.viewer
    canvas = qt_viewer.canvas

    viewer.scale_bar.visible = True
    viewer.text_overlay.visible = True
    viewer.text_overlay.text = 'test'
    viewer.scale_bar.position = 'bottom_left'
    viewer.text_overlay.position = 'bottom_left'

    vispy_scale_bar = canvas._overlay_to_visual[viewer.scale_bar][0]
    vispy_text_overlay = canvas._overlay_to_visual[viewer.text_overlay][0]

    padding = 10.0  # currently hardcoded
    y_max, x_max = canvas.size

    scale_bar_y_size = vispy_scale_bar.y_size + padding
    scale_bar_x_size = vispy_scale_bar.x_size + padding

    text_overlay_y_size = vispy_text_overlay.y_size + padding
    text_overlay_x_size = vispy_text_overlay.x_size + padding

    # check vertical tiling works on the bottom right
    viewer.scale_bar.position = 'bottom_right'
    viewer.text_overlay.position = 'bottom_right'
    canvas._update_overlay_canvas_positions()

    np.testing.assert_almost_equal(
        vispy_text_overlay.node.transform.translate[0],
        x_max - text_overlay_x_size,
        decimal=3,
    )
    np.testing.assert_almost_equal(
        vispy_text_overlay.node.transform.translate[1],
        y_max - text_overlay_y_size - scale_bar_y_size,
        decimal=3,
    )

    # move scale bar out of the way and check tiling is updated
    viewer.scale_bar.position = 'top_right'
    canvas._update_overlay_canvas_positions()
    np.testing.assert_almost_equal(
        vispy_text_overlay.node.transform.translate[0],
        x_max - text_overlay_x_size,
        decimal=3,
    )
    np.testing.assert_almost_equal(
        vispy_text_overlay.node.transform.translate[1],
        y_max - text_overlay_y_size,
        decimal=3,
    )

    # check horizontal tiling works on the top right
    viewer.text_overlay.position = 'top_right'
    canvas._update_overlay_canvas_positions()
    np.testing.assert_almost_equal(
        vispy_text_overlay.node.transform.translate[0],
        x_max - text_overlay_x_size - scale_bar_x_size,
        decimal=3,
    )
    np.testing.assert_almost_equal(
        vispy_text_overlay.node.transform.translate[1],
        0 + padding,
        decimal=3,
    )


def test_world_units_restored_after_removing_inconsistent_layer(qt_viewer):
    """Removing a units-inconsistent layer should re-enable unit-aware rendering.

    Regression test for #8771: when a layer with pixel (dimensionless) units is added to
    a viewer that already has length-unit layers, world units become inconsistent
    and unit-aware rendering is disabled.Upon removing the inconsistent layer,
    the canvas must call _update_world_units() on the next draw so the remaining compatible
    layers resume using the shared world units.
    """
    from pint import get_application_registry

    reg = get_application_registry()

    viewer = qt_viewer.viewer
    canvas = qt_viewer.canvas

    # Two images with compatible length units (um and nm share the same
    # dimensionality; consistent world units = nm, the smaller one).
    im1 = viewer.add_image(np.zeros((10, 10)), units=('um', 'um'))
    viewer.add_image(np.zeros((10, 10)), units=('nm', 'nm'))

    # Units should be consistent after adding compatible layers.
    assert viewer.layers.extent.units is not None
    vispy_im1 = canvas.layer_to_visual[im1]

    # the vispy layer received the shared world units (nm). (not just im1's own layer units (um))
    assert vispy_im1._world_units == (reg.nm, reg.nm)

    # Add layer with incompatible units (pixels)
    labels = viewer.add_labels(np.zeros((10, 10), dtype=int))

    # Units are now inconsistent across layers; _update_world_units() sets
    # world_units=None on each vispy layer, which causes the vispy layer's
    # _world_units to fall back to its own layer-local units.
    assert viewer.layers.extent.units is None
    assert vispy_im1._world_units == (
        reg.um,
        reg.um,
    )  # im1's own layer units, not (nm, nm)

    # Remove the incompatible layer.
    viewer.layers.remove(labels)

    canvas.on_draw(None)
    assert vispy_im1.world_units == (reg.nm, reg.nm)


def test_world_units_applied_to_inserted_layer_via_layerlist_event(qt_viewer):
    from pint import get_application_registry

    reg = get_application_registry()

    viewer = qt_viewer.viewer
    canvas = qt_viewer.canvas

    viewer.add_image(np.zeros((10, 10)), units=('um', 'um'))
    image_nm = viewer.add_image(np.zeros((10, 10)), units=('nm', 'nm'))

    assert viewer.layers.extent.units == (reg.nm, reg.nm)
    assert canvas.layer_to_visual[image_nm].world_units == (reg.nm, reg.nm)


def test_inserted_layer_receives_shared_world_units_when_units_unchanged(
    qt_viewer,
):
    from pint import get_application_registry

    reg = get_application_registry()

    viewer = qt_viewer.viewer
    canvas = qt_viewer.canvas

    viewer.add_image(np.zeros((10, 10)), units=('nm', 'nm'))
    image_um = viewer.add_image(np.zeros((10, 10)), units=('um', 'um'))

    assert viewer.layers.extent.units == (reg.nm, reg.nm)
    assert canvas.layer_to_visual[image_um].world_units == (reg.nm, reg.nm)



import logging

from vispy import use
from vispy.scene.visuals import Markers as BaseMarkers

logger = logging.getLogger(__name__)

try:
    use(gl='gl+')
except RuntimeError:
    logger.warning(
        'Could not use gl+ for instanced rendering of points.'
        'Falling back to GL point rendering.'
    )
    rendering_method = 'points'
else:
    rendering_method = 'instanced'


class Markers(BaseMarkers):
    def __init__(self, *args, method=rendering_method, **kwargs) -> None:
        super().__init__(*args, method=method, **kwargs)

    def _compute_bounds(self, axis, view):
        # needed for entering 3D rendering mode when a points
        # layer is invisible and the self._data property is None
        if self._data is None:
            return None
        pos = self._data['a_position']
        if pos is None or pos.size == 0:
            return None
        if pos.shape[1] > axis:
            return (pos[:, axis].min(), pos[:, axis].max())

        return (0, 0)



from typing import Any

from vispy.scene.visuals import Text as BaseText

from napari._vispy.utils.qt_font import QtFontManager
from napari._vispy.utils.text import (
    get_text_width_height,
)

# Global Qt-based font manager instance shared across all Text visuals
_qt_font_manager = None

_FONT_FAMILY = 'OpenSans'


def get_qt_font_manager(method: str = 'cpu') -> QtFontManager:
    """Get or create the global Qt font manager instance.

    Parameters
    ----------
    method : str, optional
        Rendering method ('cpu' or 'gpu'). Default is 'cpu'.

    Returns
    -------
    QtFontManager
        The global Qt font manager instance.
    """
    global _qt_font_manager
    if _qt_font_manager is None:
        _qt_font_manager = QtFontManager(method=method)
    return _qt_font_manager


class Text(BaseText):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # If using Qt fonts, pass the Qt font manager to the base class
        if 'face' not in kwargs:
            kwargs['face'] = _FONT_FAMILY
        super().__init__(*args, **kwargs)

    def get_width_height(self) -> tuple[float, float]:
        width, height = get_text_width_height(self)
        return width * self._vispy_dpi_ratio, height * self._vispy_dpi_ratio

    @property
    def font_size(self) -> float:
        return self._font_size * self.dpi_ratio

    @font_size.setter
    def font_size(self, size: float) -> None:
        adjusted_font_size = size / self.dpi_ratio
        self._font_size = max(0.0, adjusted_font_size)
        self.update()

    @property
    def _vispy_dpi_ratio(self) -> float:
        # while 72 is considered the base dpi in vispy, most software assumes
        # 96 as reference dpi, resulting in mismatch between gui and canvas
        # when font sizes are the same. We need to account for this when
        # we calculate things externally from vispy and
        return 96 / 72

    @property
    def dpi_ratio(self) -> float:
        dpi = self.transforms.dpi or 96
        return dpi / 96



import dask.array as da
import numpy as np
import pytest
import zarr

from napari.layers import Image
from napari.layers.base._base_constants import Blending
from napari.layers.utils.stack_utils import (
    images_to_stack,
    merge_rgb,
    slice_from_axis,
    split_channels,
    split_rgb,
    stack_to_images,
)
from napari.utils.transforms import Affine


def test_stack_to_images_basic():
    """Test that a 2 channel zcyx stack is split into 2 image layers"""
    data = np.random.randint(0, 100, (10, 2, 128, 128))
    stack = Image(data)
    images = stack_to_images(stack, 1, colormap=None)

    assert isinstance(images, list)
    assert images[0].colormap.name == 'magenta'
    assert len(images) == 2

    for i in images:
        assert type(stack) is type(i)
        assert i.data.shape == (10, 128, 128)


def test_stack_to_images_multiscale():
    """Test that a 3 channel multiscale image returns 3 multiscale images."""
    data = []
    data.append(np.zeros((3, 128, 128)))
    data.append(np.zeros((3, 64, 64)))
    data.append(np.zeros((3, 32, 32)))
    data.append(np.zeros((3, 16, 16)))

    stack = Image(data)
    images = stack_to_images(stack, 0)

    assert len(images) == 3
    assert len(images[0].data) == 4
    assert images[0].data[-1].shape[-1] == 16
    assert images[1].data[-1].shape[-1] == 16
    assert images[2].data[-1].shape[-1] == 16


def test_stack_to_images_rgb():
    """Test 3 channel RGB image (channel axis = -1) into single channels."""
    data = np.random.randint(0, 100, (10, 128, 128, 3))
    stack = Image(data)
    images = stack_to_images(stack, -1, colormap=None)

    assert isinstance(images, list)
    assert len(images) == 3

    for i in images:
        assert type(stack) is type(i)
        assert i.data.shape == (10, 128, 128)
        assert i.scale.shape == (3,)
        assert i.rgb is False


def test_stack_to_images_4_channels():
    """Test 4x128x128 stack is split into 4 channels w/ colormap keyword"""
    data = np.random.randint(0, 100, (4, 128, 128))
    stack = Image(data)
    images = stack_to_images(stack, 0, colormap=['red', 'blue'])

    assert isinstance(images, list)
    assert len(images) == 4
    assert images[-2].colormap.name == 'red'
    for i in images:
        assert type(stack) is type(i)
        assert i.data.shape == (128, 128)


def test_stack_to_images_0_rgb():
    """Split RGB along the first axis (z or t) so the images remain rgb"""
    data = np.random.randint(0, 100, (10, 128, 128, 3))
    stack = Image(data)
    images = stack_to_images(stack, 0, colormap=None)

    assert isinstance(images, list)
    assert len(images) == 10
    for i in images:
        assert i.rgb
        assert type(stack) is type(i)
        assert i.data.shape == (128, 128, 3)


def test_stack_to_images_1_channel():
    """Split when only one channel"""
    data = np.random.randint(0, 100, (10, 1, 128, 128))
    stack = Image(data)
    images = stack_to_images(stack, 1, colormap=['magma'])

    assert isinstance(images, list)
    assert len(images) == 1
    for i in images:
        assert i.rgb is False
        assert type(stack) is type(i)
        assert i.data.shape == (10, 128, 128)


def test_images_to_stack_with_scale():
    """Test that 3-Image list is combined to stack with scale and translate."""
    images = [
        Image(np.random.randint(0, 255, (10, 128, 128))) for _ in range(3)
    ]

    stack = images_to_stack(
        images, 1, colormap='green', scale=(3, 1, 1, 1), translate=(1, 0, 2, 3)
    )

    assert isinstance(stack, Image)
    assert stack.data.shape == (10, 3, 128, 128)
    assert stack.colormap.name == 'green'
    assert list(stack.scale) == [3, 1, 1, 1]
    assert list(stack.translate) == [1, 0, 2, 3]


def test_images_to_stack_none_scale():
    """Test combining images using scale & translate from 1st image in list"""
    images = [
        Image(
            np.random.randint(0, 255, (10, 128, 128)),
            scale=(4, 1, 1),
            translate=(0, -1, 2),
        )
        for _ in range(3)
    ]

    stack = images_to_stack(images, 1, colormap='green')

    assert isinstance(stack, Image)
    assert stack.data.shape == (10, 3, 128, 128)
    assert stack.colormap.name == 'green'
    assert list(stack.scale) == [4, 1, 1, 1]
    assert list(stack.translate) == [0, 0, -1, 2]


def test_images_to_stack_multiscale():
    data = np.zeros((100, 100))
    images = [
        Image([data, data[::2, ::2]]),
        Image([data, data[::2, ::2]]),
    ]
    assert images[0].multiscale is True
    assert images[1].multiscale is True

    stack = images_to_stack(images, 0, colormap='green')

    assert isinstance(stack, Image)
    assert stack.multiscale is True
    assert stack.data.shape == (2, 100, 100)


def test_split_and_merge_rgb():
    """Test merging 3 images with RGB colormaps into single RGB image."""
    # Make an RGB
    data = np.random.randint(0, 100, (10, 128, 128, 3))
    stack = Image(data)
    assert stack.rgb is True

    # split the RGB into 3 images
    images = split_rgb(stack)
    assert len(images) == 3
    colormaps = {image.colormap.name for image in images}
    assert colormaps == {'red', 'green', 'blue'}

    # merge the 3 images back into an RGB
    rgb_image = merge_rgb(images)
    assert rgb_image.rgb is True


def test_split_and_merge_rgba():
    """Test splitting a rgba image into channels and then re-merging."""
    data = np.random.randint(0, 100, (10, 128, 128, 4))
    # set channels to distinct values to aid in confirming re-merging image
    data[0] = 1
    data[1] = 2
    data[2] = 3
    data[3] = 4
    stack = Image(data)
    assert stack.rgb is True

    # split the rgb into 4 images
    images = split_rgb(stack, with_alpha=True)
    assert len(images) == 4
    colormaps = {image.colormap.name for image in images}
    # gray should be assigned to alpha channel
    assert colormaps == {'red', 'green', 'blue', 'gray'}

    # merge the 4 images back into a rgba image
    rgb_image = merge_rgb(images)
    assert rgb_image.rgb is True
    assert rgb_image.data.shape[-1] == 4
    # confirm that channel are assigned correctly
    assert (rgb_image.data[0] == 1).all()
    assert (rgb_image.data[1] == 2).all()
    assert (rgb_image.data[2] == 3).all()
    assert (rgb_image.data[3] == 4).all()


@pytest.mark.parametrize(
    'stack_blending', [blending.value for blending in Blending]
)
def test_split_rgb_blending(stack_blending):
    """Test blending settings on a split RGB image."""
    # Make an RGB
    data = np.random.randint(0, 100, (10, 128, 128, 3))
    stack = Image(data)
    stack.blending = stack_blending

    # split the RGB into 3 images
    images = split_rgb(stack)
    blendings = [image.blending for image in images]
    assert blendings == [stack_blending, 'additive', 'additive']


@pytest.mark.parametrize(
    'stack_blending', [blending.value for blending in Blending]
)
def test_split_rgba_blending(stack_blending):
    """Test blending settings on a split RGBA image."""
    # Make an RGBA
    data = np.random.randint(0, 100, (10, 128, 128, 4))
    stack = Image(data)
    stack.blending = stack_blending

    # split the rgb into 4 images
    images = split_rgb(stack, with_alpha=True)
    blendings = [image.blending for image in images]
    # multiplicative should be assigned to alpha channel
    assert blendings == [
        stack_blending,
        'additive',
        'additive',
        'multiplicative',
    ]


@pytest.fixture(
    params=[
        {
            'rgb': None,
            'colormap': None,
            'contrast_limits': None,
            'gamma': 1,
            'interpolation': 'nearest',
            'rendering': 'mip',
            'iso_threshold': 0.5,
            'attenuation': 0.5,
            'name': None,
            'metadata': None,
            'scale': None,
            'translate': None,
            'opacity': 1,
            'blending': None,
            'visible': True,
            'multiscale': None,
            'rotate': None,
            'affine': None,
        },
        {
            'rgb': None,
            'colormap': None,
            'rendering': 'mip',
            'attenuation': 0.5,
            'metadata': None,
            'scale': None,
            'opacity': 1,
            'visible': True,
            'multiscale': None,
        },
        {},
    ],
    ids=['full-kwargs', 'partial-kwargs', 'empty-kwargs'],
)
def kwargs(request):
    return dict(request.param)


def test_split_channels(kwargs):
    """Test split_channels with shape (3,128,128) expecting 3 (128,128)"""
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, _meta, _ in result_list:
        assert d.shape == (128, 128)


def test_split_channels_multiscale(kwargs):
    """Test split_channels with multiscale expecting List[LayerData]"""
    data = []
    data.append(np.zeros((3, 128, 128)))
    data.append(np.zeros((3, 64, 64)))
    data.append(np.zeros((3, 32, 32)))
    data.append(np.zeros((3, 16, 16)))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for ds, m, _ in result_list:
        assert m['multiscale'] is True
        assert ds[0].shape == (128, 128)
        assert ds[1].shape == (64, 64)
        assert ds[2].shape == (32, 32)
        assert ds[3].shape == (16, 16)


def test_split_channels_blending(kwargs):
    """Test split_channels with shape (3,128,128) expecting 3 (128,128)"""
    kwargs['blending'] = 'translucent'
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, meta, _ in result_list:
        assert d.shape == (128, 128)
        assert meta['blending'] == 'translucent'


def test_split_channels_missing_keywords():
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0)

    assert len(result_list) == 3
    for chan, layer in enumerate(result_list):
        assert layer[0].shape == (128, 128)
        assert (
            layer[1]['blending'] == 'translucent_no_depth'
            if chan == 0
            else 'additive'
        )


def test_split_channels_affine_nparray(kwargs):
    kwargs['affine'] = np.eye(3)
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, meta, _ in result_list:
        assert d.shape == (128, 128)
        assert np.array_equal(meta['affine'], np.eye(3))


def test_split_channels_affine_napari(kwargs):
    kwargs['affine'] = Affine(affine_matrix=np.eye(3))
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, meta, _ in result_list:
        assert d.shape == (128, 128)
        assert np.array_equal(meta['affine'].affine_matrix, np.eye(3))


def test_split_channels_multi_affine_napari(kwargs):
    kwargs['affine'] = [
        Affine(scale=[1, 1]),
        Affine(scale=[2, 2]),
        Affine(scale=[3, 3]),
    ]

    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for idx, result_data in enumerate(result_list):
        d, meta, _ = result_data
        assert d.shape == (128, 128)
        assert np.array_equal(
            meta['affine'].affine_matrix,
            Affine(scale=[idx + 1, idx + 1]).affine_matrix,
        )


@pytest.mark.parametrize(
    ('input_array', 'expected_type'),
    [
        (np.zeros((10, 20)), np.ndarray),
        (da.zeros((10, 20)), da.Array),
        (zarr.zeros((10, 20)), da.Array),
    ],
    ids=['numpy', 'dask', 'zarr'],
)
def test_images_to_stack_lazy_arrays(input_array, expected_type):
    """Test that images_to_stack handles numpy, dask, and zarr arrays correctly."""
    data = input_array
    images = [Image(data) for _ in range(3)]
    stack = images_to_stack(images)

    assert not stack.multiscale
    assert isinstance(stack.data, expected_type)
    assert stack.data.shape[1:] == input_array.shape


@pytest.mark.parametrize(
    ('input_array', 'expected_type'),
    [
        (np.zeros((10, 20)), np.ndarray),
        (da.zeros((10, 20)), da.Array),
        (zarr.zeros((10, 20)), da.Array),
    ],
    ids=['numpy', 'dask', 'zarr'],
)
def test_images_to_stack_lazy_multiscale_arrays(input_array, expected_type):
    """Test stacking of multiscale numpy, dask, and zarr arrays."""
    # Slicing zarr array returns numpy array, so we need to re-wrap in zarr
    if isinstance(input_array, zarr.Array):
        data = [input_array, zarr.array(input_array[::2, ::2])]
    else:
        data = [input_array, input_array[::2, ::2]]

    images = [Image(data) for _ in range(3)]
    stack = images_to_stack(images)

    assert stack.multiscale
    assert isinstance(stack.data[0], expected_type)
    assert stack.data[0].shape[1:] == input_array.shape


@pytest.mark.parametrize(
    ('array_type', 'expected_result_type'),
    [
        ('numpy', np.ndarray),
        ('dask', da.Array),
        ('zarr', da.Array),
    ],
)
def test_slice_from_axis_different_array_types(
    array_type, expected_result_type
):
    """Test slice_from_axis with numpy, dask, and zarr arrays."""
    # Create test data
    if array_type == 'numpy':
        data = np.zeros((3, 4, 4))
    elif array_type == 'dask':
        data = da.zeros((3, 4, 4))
    elif array_type == 'zarr':
        data = zarr.zeros((3, 4, 4))

    axis, element = 1, 2
    expected_shape = (3, 4)
    result = slice_from_axis(data, axis=axis, element=element)

    # Check result type and shape
    assert isinstance(result, expected_result_type)
    assert result.shape == expected_shape

    # Check result values - compare with expected slice
    result_computed = (
        result.compute() if hasattr(result, 'compute') else result
    )
    expected = data[:, element, :]
    np.testing.assert_array_equal(result_computed, expected)



import warnings
from dataclasses import dataclass
from typing import Any

import numpy as np
from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

from napari.utils.translations import trans


@dataclass(eq=False)
class ColorCycle:
    """A dataclass to hold a color cycle for the fallback_colors
    in the CategoricalColormap

    Attributes
    ----------
    values : np.ndarray
        The (Nx4) color array of all colors contained in the color cycle.
    current_index : int
        The index of the current color within the values array.
    """

    values: np.ndarray
    current_index: int = 0

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        validate = core_schema.no_info_plain_validator_function(
            cls.validate_type
        )
        json_serialize = core_schema.plain_serializer_function_ser_schema(
            lambda v: {
                'values': v.values.tolist()
                if isinstance(v.values, np.ndarray)
                else v,
                'current_index': v.current_index,
            },
            when_used='json',
        )
        return core_schema.json_or_python_schema(
            json_schema=validate,
            python_schema=validate,
            serialization=json_serialize,
        )

    @classmethod
    def validate_type(cls, val):
        # turn a generic dict into object
        if isinstance(val, dict):
            return _coerce_colorcycle_from_dict(val)
        if isinstance(val, ColorCycle):
            return val

        return _coerce_colorcycle_from_colors(val)

    def __eq__(self, other):
        if isinstance(other, ColorCycle):
            eq = np.array_equal(self.values, other.values)
        else:
            eq = False
        return eq

    def __next__(self):
        val = self.current_color()
        self.current_index += 1
        self.current_index %= len(self.values)
        return val

    def __iter__(self):
        return self

    def current_color(self) -> np.ndarray:
        return self.values[self.current_index]

    @property
    def cycle(self):
        warnings.warn(
            'ColorCycle.cycle is deprecated and will be removed in 0.8.0. '
            'To iterate colors, use next(ColorCycle) directly.',
            stacklevel=2,
        )
        return self


def _coerce_colorcycle_from_dict(
    val: dict[str, str | list | np.ndarray | int],
) -> ColorCycle:
    # avoid circular import
    from napari.layers.utils.color_transformations import (
        transform_color,
    )

    # validate values
    color_values = val.get('values')
    if color_values is None:
        raise ValueError(
            trans._('ColorCycle requires a values argument', deferred=True)
        )

    transformed_color_values = transform_color(color_values)

    current_index = val.get('current_index', 0)
    if not isinstance(current_index, int):
        raise TypeError('ColorCycle current_index must be an integer')
    if current_index > len(transformed_color_values) - 1:
        raise ValueError('ColorCycle current_index must be < len(values)')

    return ColorCycle(
        values=transformed_color_values,
        current_index=current_index,
    )


def _coerce_colorcycle_from_colors(
    val: str | list | np.ndarray,
) -> ColorCycle:
    # avoid circular import
    from napari.layers.utils.color_transformations import (
        transform_color_cycle,
    )

    if isinstance(val, str):
        val = [val]

    transformed_color_values = transform_color_cycle(
        color_cycle=val,
        elem_name='color_cycle',
        default='white',
    )
    return ColorCycle(values=transformed_color_values)


def compare_colormap_dicts(cmap_1, cmap_2):
    if len(cmap_1) != len(cmap_2):
        return False
    for k, v in cmap_1.items():
        if k not in cmap_2:
            return False
        if not np.allclose(v, cmap_2[k]):
            return False
    return True



import numpy as np

from napari.layers.image import Image, _image_key_bindings as key_bindings


def test_auto_contrast_once_sets_clims():
    data = np.array([[0, 10]])
    layer = Image(data)
    # change to something else first
    layer.contrast_limits = (1, 2)
    key_bindings.auto_contrast_once(layer)
    assert layer.contrast_limits == [0.0, 10.0]



import numpy as np
from vispy.scene.visuals import Compound, Line
from vispy.visuals.text.text import FontManager

from napari._vispy.visuals.text import Text


class ScaleBar(Compound):
    """Scale bar visual with text and line components.

    Layout of Scale Bar Elements (from top to bottom):
    - Padding
    - Text
    - Gap between text and line
    - Scale line (with optional ticks)
    - Padding
    """

    def __init__(
        self,
        font_manager: FontManager | None = None,
        font_family: str = 'OpenSans',
    ) -> None:
        # Layout constants
        self.PADDING = 6  # Space around the entire scale bar
        self.TICK_LENGTH = 11  # Height of tick marks (odd numbers look better)

        # Line geometry: main line + optional tick marks
        self._line_data = np.array(
            [
                [-1, 0],  # Left end of main line
                [1, 0],  # Right end of main line
                [-1, -1],  # Left tick mark (bottom)
                [-1, 1],  # Left tick mark (top)
                [1, -1],  # Right tick mark (bottom)
                [1, 1],  # Right tick mark (top)
            ]
        )

        self._color = (1, 1, 1, 1)

        self.text = Text(
            text='1px',
            pos=[0.5, 0.5],
            anchor_x='center',
            anchor_y='bottom',
            font_size=10,
            face=font_family,
            font_manager=font_manager,
        )
        self.line = Line(
            connect='segments', method='gl', width=3, antialias=True
        )
        # order matters (last is drawn on top)
        super().__init__([self.text, self.line])

    def _calculate_layout(self, length: float) -> dict:
        """Calculate all layout dimensions and positions."""
        # Text dimensions
        text_width, text_height = self.text.get_width_height()

        # ceil so there's no flickering due to slight differences between
        # the heigh of the various digits
        text_height = np.ceil(text_height)

        width = max(
            length + self.line.width,
            text_width,
        )
        height = text_height + self.TICK_LENGTH

        line_center_y = text_height + (self.TICK_LENGTH / 2)

        return {
            'width': width,
            'height': height,
            'line_center_y': line_center_y,
        }

    def set_data(self, *, length, color, ticks, font_size):
        """Update scale bar with new dimensions and styling."""
        # font size need to be set first cause layout calculations depend on it
        self.text.font_size = font_size

        layout = self._calculate_layout(length)

        # Choose line data based on whether ticks are enabled
        line_data = self._line_data if ticks else self._line_data[:2]

        # Position and scale the line
        self.line.set_data(
            pos=line_data * (length / 2, self.TICK_LENGTH / 2)
            + (
                layout['width'] / 2,  # Center horizontally
                layout['line_center_y'],  # Position vertically
            ),
            color=color,
        )

        # Position the text
        self.text.pos = layout['width'] / 2, 0
        self.text.color = color

        # Return dimensions for the overlay system
        # Extra padding needed for proper canvas positioning (not sure why padding is needed here, ugh)
        return layout['width'], layout['height']



"""Text label model."""

from napari.components._viewer_constants import CanvasPosition
from napari.components.overlays.base import CanvasOverlay
from napari.utils.color import ColorValue


class _BaseTextOverlay(CanvasOverlay):
    """Label model to display arbitrary text in the canvas

    Attributes
    ----------
    color : np.ndarray
        A (4,) color array of the text overlay.
    font_size : float
        The font size (in points) of the text.
    position : CanvasPosition
        The position of the overlay in the canvas.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    color: ColorValue | None = None
    font_size: float = 10


class TextOverlay(_BaseTextOverlay):
    """Label model to display arbitrary text in the canvas

    Attributes
    ----------
    color : np.ndarray
        A (4,) color array of the text overlay.
    font_size : float
        The font size (in points) of the text.
    text : str
        Text to be displayed in the canvas.
    position : CanvasPosition
        The position of the overlay in the canvas.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    text: str = ''


class LayerNameOverlay(_BaseTextOverlay):
    """Label model to display layer name text in the canvas

    Attributes
    ----------
    color : np.ndarray
        A (4,) color array of the text overlay.
    font_size : float
        The font size (in points) of the text.
    position : CanvasPosition
        The position of the overlay in the canvas.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    position: CanvasPosition = CanvasPosition.TOP_LEFT


class CurrentSliceOverlay(_BaseTextOverlay):
    """Label model to display the current dims slice in the canvas

    Attributes
    ----------
    color : np.ndarray
        A (4,) color array of the text overlay.
    font_size : float
        The font size (in points) of the text.
    position : CanvasPosition
        The position of the overlay in the canvas.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """



from typing import (
    TYPE_CHECKING,
)

import numpy as np
from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

if TYPE_CHECKING:
    from decimal import Decimal

    Number = int | float | Decimal

# In numpy 2, the semantics of the copy argument in np.array changed
# so that copy=False errors if a copy is needed:
# https://numpy.org/devdocs/numpy_2_0_migration_guide.html#adapting-to-changes-in-the-copy-keyword
#
# In numpy 1, copy=False meant that a copy was avoided unless necessary,
# but would not error.
#
# In most usage like this use np.asarray instead, but sometimes we need
# to use some of the unique arguments of np.array (e.g. ndmin).
#
# This solution assumes numpy 1 by default, and switches to the numpy 2
# value for any release of numpy 2 on PyPI (including betas and RCs).
copy_if_needed: bool | None = False
if np.lib.NumpyVersion(np.__version__) >= '2.0.0b1':
    copy_if_needed = None


class Array(np.ndarray):
    def __class_getitem__(cls, t):
        return type('Array', (Array,), {'__dtype__': t})

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source, handler: GetCoreSchemaHandler
    ):
        return core_schema.no_info_after_validator_function(
            cls.validate_type, core_schema.any_schema()
        )

    @classmethod
    def validate_type(cls, val):
        dtype = getattr(cls, '__dtype__', None)
        if isinstance(dtype, tuple):
            dtype, shape = dtype
        else:
            shape = ()

        result = np.array(
            val, dtype=dtype, copy=copy_if_needed, ndmin=len(shape)
        )

        if any(
            (shape[i] != -1 and shape[i] != result.shape[i])
            for i in range(len(shape))
        ):
            result = result.reshape(shape)
        return result


class NotEqual:
    def __init__(self, ne):
        self.ne = ne

    def __get_pydantic_core_schema__(
        self, source, handler: GetCoreSchemaHandler
    ):
        return core_schema.no_info_after_validator_function(
            self._validate, handler(source)
        )

    def _validate(self, v):
        if v == self.ne:
            raise ValueError(f'value must not be equal to {self.ne}')
        return v



import importlib
from collections import defaultdict
from itertools import product
from unittest.mock import patch

import numpy as np
import numpy.testing as npt
import pytest
from pydantic import ValidationError

from napari.utils.color import ColorArray
from napari.utils.colormap_backend import ColormapBackend, set_backend
from napari.utils.colormaps import (
    Colormap,
    _accelerated_cmap,
    colormap,
)
from napari.utils.colormaps.colormap import (
    CyclicLabelColormap,
    DirectLabelColormap,
    LabelColormapBase,
    _normalize_label_colormap,
)
from napari.utils.colormaps.colormap_utils import label_colormap

backends = [ColormapBackend.pure_python]
if _accelerated_cmap.numba is not None:
    backends.append(ColormapBackend.numba)
if _accelerated_cmap.partsegcore_mapping is not None:
    backends.append(ColormapBackend.partsegcore)


@pytest.fixture(autouse=True, params=backends)
def colormap_backend(request):
    """Fixture to set the colormap backend for each test."""
    backend = request.param
    set_backend(backend)
    yield backend
    # Reset the backend after the test
    set_backend(ColormapBackend.fastest_available)


def test_linear_colormap():
    """Test a linear colormap."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    cmap = Colormap(colors, name='testing')

    assert cmap.name == 'testing'
    assert cmap.interpolation == 'linear'
    assert len(cmap.controls) == len(colors)
    np.testing.assert_almost_equal(cmap.colors, colors)
    np.testing.assert_almost_equal(cmap.map([0.75]), [[0, 0.5, 0.5, 1]])


def test_linear_colormap_with_control_points():
    """Test a linear colormap with control points."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    cmap = Colormap(colors, name='testing', controls=[0, 0.75, 1])

    assert cmap.name == 'testing'
    assert cmap.interpolation == 'linear'
    assert len(cmap.controls) == len(colors)
    np.testing.assert_almost_equal(cmap.colors, colors)
    np.testing.assert_almost_equal(cmap.map([0.75]), [[0, 1, 0, 1]])


def test_non_ascending_control_points():
    """Test non ascending control points raises an error."""
    colors = np.array(
        [[0, 0, 0, 1], [0, 0.5, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]]
    )
    with pytest.raises(
        ValidationError, match='need to be sorted in ascending order'
    ):
        Colormap(colors, name='testing', controls=[0, 0.75, 0.25, 1])


def test_wrong_number_control_points():
    """Test wrong number of control points raises an error."""
    colors = np.array(
        [[0, 0, 0, 1], [0, 0.5, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]]
    )
    with pytest.raises(
        ValidationError, match='Wrong number of control points'
    ):
        Colormap(colors, name='testing', controls=[0, 0.75, 1])


def test_wrong_start_control_point():
    """Test wrong start of control points raises an error."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    with pytest.raises(
        ValidationError, match=r'must start with 0.0 and end with 1.0'
    ):
        Colormap(colors, name='testing', controls=[0.1, 0.75, 1])


def test_wrong_end_control_point():
    """Test wrong end of control points raises an error."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    with pytest.raises(
        ValidationError, match=r'must start with 0.0 and end with 1.0'
    ):
        Colormap(colors, name='testing', controls=[0, 0.75, 0.9])


def test_binned_colormap():
    """Test a binned colormap."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    cmap = Colormap(colors, name='testing', interpolation='zero')

    assert cmap.name == 'testing'
    assert cmap.interpolation == 'zero'
    assert len(cmap.controls) == len(colors) + 1
    np.testing.assert_almost_equal(cmap.colors, colors)
    np.testing.assert_almost_equal(cmap.map([0.4]), [[0, 1, 0, 1]])


def test_binned_colormap_with_control_points():
    """Test a binned with control points."""
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    cmap = Colormap(
        colors,
        name='testing',
        interpolation='zero',
        controls=[0, 0.2, 0.3, 1],
    )

    assert cmap.name == 'testing'
    assert cmap.interpolation == 'zero'
    assert len(cmap.controls) == len(colors) + 1
    np.testing.assert_almost_equal(cmap.colors, colors)
    np.testing.assert_almost_equal(cmap.map([0.4]), [[0, 0, 1, 1]])


def test_colormap_equality():
    colors = np.array([[0, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1]])
    cmap_1 = Colormap(colors, name='testing', controls=[0, 0.75, 1])
    cmap_2 = Colormap(colors, name='testing', controls=[0, 0.75, 1])
    cmap_3 = Colormap(colors, name='testing', controls=[0, 0.25, 1])
    assert cmap_1 == cmap_2
    assert cmap_1 != cmap_3


def test_colormap_recreate():
    c_map = Colormap('black')
    Colormap(**c_map.model_dump())


@pytest.mark.parametrize('ndim', range(1, 5))
def test_mapped_shape(ndim):
    np.random.seed(0)
    img = np.random.random((5,) * ndim)
    cmap = Colormap(colors=['red'])
    mapped = cmap.map(img)
    assert mapped.shape == img.shape + (4,)


@pytest.mark.parametrize(
    ('num', 'dtype'), [(40, np.uint8), (1000, np.uint16), (80000, np.float32)]
)
def test_minimum_dtype_for_labels(num, dtype):
    assert _accelerated_cmap.minimum_dtype_for_labels(num) == dtype


@pytest.fixture
def _disable_jit(monkeypatch):
    """Fixture to temporarily disable numba JIT during testing.

    This helps to measure coverage and in debugging. *However*, reloading a
    module can cause issues with object instance / class relationships, so
    the `_accelerated_cmap` module should be as small as possible and contain
    no class definitions, only functions.
    """
    pytest.importorskip('numba')

    with patch('numba.core.config.DISABLE_JIT', True):
        importlib.reload(_accelerated_cmap)
        yield
    importlib.reload(_accelerated_cmap)  # revert to original state


@pytest.mark.parametrize(('num', 'dtype'), [(40, np.uint8), (1000, np.uint16)])
@pytest.mark.usefixtures('_disable_jit')
def test_cast_labels_to_minimum_type_auto(num: int, dtype, monkeypatch):
    cmap = label_colormap(num)
    data = np.zeros(3, dtype=np.uint32)
    data[1] = 10
    data[2] = 10**6 + 5
    cast_arr = colormap._cast_labels_data_to_texture_dtype_auto(data, cmap)
    assert cast_arr.dtype == dtype
    assert cast_arr[0] == 0
    assert cast_arr[1] == 10
    assert cast_arr[2] == 10**6 % num + 5


@pytest.fixture
def direct_label_colormap():
    return DirectLabelColormap(
        color_dict={
            0: np.array([0, 0, 0, 0]),
            1: np.array([1, 0, 0, 1]),
            2: np.array([0, 1, 0, 1]),
            3: np.array([0, 0, 1, 1]),
            12: np.array([0, 0, 1, 1]),
            None: np.array([1, 1, 1, 1]),
        },
    )


def test_direct_label_colormap_simple(direct_label_colormap):
    np.testing.assert_array_equal(
        direct_label_colormap.map([0, 2, 7]),
        np.array([[0, 0, 0, 0], [0, 1, 0, 1], [1, 1, 1, 1]]),
    )
    assert direct_label_colormap._num_unique_colors == 5

    (
        label_mapping,
        color_dict,
    ) = direct_label_colormap._values_mapping_to_minimum_values_set()

    assert len(label_mapping) == 6
    assert len(color_dict) == 5
    assert label_mapping[None] == _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE
    assert label_mapping[12] == label_mapping[3]
    np.testing.assert_array_equal(
        color_dict[label_mapping[0]], direct_label_colormap.color_dict[0]
    )
    np.testing.assert_array_equal(
        color_dict[0], direct_label_colormap.color_dict[None]
    )


def test_direct_label_colormap_selection(direct_label_colormap):
    direct_label_colormap.selection = 2
    direct_label_colormap.use_selection = True

    np.testing.assert_array_equal(
        direct_label_colormap.map([0, 2, 7]),
        np.array([[0, 0, 0, 0], [0, 1, 0, 1], [0, 0, 0, 0]]),
    )

    (
        label_mapping,
        color_dict,
    ) = direct_label_colormap._values_mapping_to_minimum_values_set()

    assert len(label_mapping) == 2
    assert len(color_dict) == 2


@pytest.mark.usefixtures('_disable_jit')
def test_cast_direct_labels_to_minimum_type(direct_label_colormap):
    data = np.arange(15, dtype=np.uint32)
    cast = _accelerated_cmap.labels_raw_to_texture_direct(
        data, direct_label_colormap
    )
    label_mapping = (
        direct_label_colormap._values_mapping_to_minimum_values_set()[0]
    )
    assert cast.dtype == np.uint8
    np.testing.assert_array_equal(
        cast,
        np.array(
            [
                label_mapping[0],
                label_mapping[1],
                label_mapping[2],
                label_mapping[3],
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                label_mapping[3],
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
                _accelerated_cmap.MAPPING_OF_UNKNOWN_VALUE,
            ]
        ),
    )


@pytest.mark.parametrize(
    ('num', 'dtype'), [(40, np.uint8), (1000, np.uint16), (80000, np.float32)]
)
@pytest.mark.usefixtures('_disable_jit')
def test_test_cast_direct_labels_to_minimum_type_no_jit(num, dtype):
    cmap = DirectLabelColormap(
        color_dict={
            None: np.array([1, 1, 1, 1]),
            **{
                k: np.array([*v, 1])
                for k, v in zip(
                    range(num),
                    product(np.linspace(0, 1, num=256), repeat=3),
                    strict=False,
                )
            },
        },
    )
    cmap.color_dict[None] = np.array([1, 1, 1, 1])
    data = np.arange(10, dtype=np.uint32)
    data[2] = 80005
    cast = _accelerated_cmap.labels_raw_to_texture_direct(data, cmap)
    assert cast.dtype == dtype


def test_zero_preserving_modulo_naive_vs_numba():
    pytest.importorskip('numba')

    data = np.arange(1000, dtype=np.uint32)
    res1 = _accelerated_cmap.zero_preserving_modulo_numpy(data, 49, np.uint8)
    res2 = _accelerated_cmap.zero_preserving_modulo(data, 49, np.uint8)
    npt.assert_array_equal(res1, res2)


def test_zero_preserving_modulo_naive_vs_partseg():
    pytest.importorskip('PartSegCore_compiled_backend.napari_mapping')
    data = np.arange(1000, dtype=np.uint32)
    res1 = _accelerated_cmap.zero_preserving_modulo_numpy(data, 49, np.uint8)
    res2 = _accelerated_cmap.zero_preserving_modulo_partsegcore(
        data, 49, np.uint8
    )
    npt.assert_array_equal(res1, res2)


@pytest.mark.parametrize(
    'dtype', [np.uint8, np.uint16, np.int8, np.int16, np.float32, np.float64]
)
def test_label_colormap_map_with_uint8_values(dtype):
    cmap = colormap.CyclicLabelColormap(
        colors=ColorArray(np.array([[0, 0, 0, 0], [1, 0, 0, 1], [0, 1, 0, 1]]))
    )
    values = np.array([0, 1, 2], dtype=dtype)
    expected = np.array([[0, 0, 0, 0], [1, 0, 0, 1], [0, 1, 0, 1]])
    npt.assert_array_equal(cmap.map(values), expected)


@pytest.mark.parametrize('selection', [1, -1])
@pytest.mark.parametrize('dtype', [np.int8, np.int16, np.int32, np.int64])
def test_label_colormap_map_with_selection(selection, dtype):
    cmap = colormap.CyclicLabelColormap(
        colors=ColorArray(
            np.array([[0, 0, 0, 0], [1, 0, 0, 1], [0, 1, 0, 1]])
        ),
        use_selection=True,
        selection=selection,
    )
    values = np.array([0, selection, 2], dtype=np.int8)
    expected = np.array([[0, 0, 0, 0], [1, 0, 0, 1], [0, 0, 0, 0]])
    npt.assert_array_equal(cmap.map(values), expected)


@pytest.mark.parametrize('background', [1, -1])
@pytest.mark.parametrize('dtype', [np.int8, np.int16, np.int32, np.int64])
def test_label_colormap_map_with_background(background, dtype):
    cmap = colormap.CyclicLabelColormap(
        colors=ColorArray(
            np.array([[0, 0, 0, 0], [1, 0, 0, 1], [0, 1, 0, 1]])
        ),
        background_value=background,
    )
    values = np.array([3, background, 2], dtype=dtype)
    expected = np.array([[1, 0, 0, 1], [0, 0, 0, 0], [0, 1, 0, 1]])
    npt.assert_array_equal(cmap.map(values), expected)


@pytest.mark.parametrize('dtype', [np.uint8, np.uint16])
def test_label_colormap_using_cache(dtype, monkeypatch):
    cmap = colormap.CyclicLabelColormap(
        colors=ColorArray(np.array([[0, 0, 0, 0], [1, 0, 0, 1], [0, 1, 0, 1]]))
    )
    values = np.array([0, 1, 2], dtype=dtype)
    expected = np.array([[0, 0, 0, 0], [1, 0, 0, 1], [0, 1, 0, 1]])
    map1 = cmap.map(values)
    npt.assert_array_equal(map1, expected)
    monkeypatch.setattr(
        _accelerated_cmap, 'zero_preserving_modulo_numpy', None
    )
    map2 = cmap.map(values)
    npt.assert_array_equal(map1, map2)


@pytest.mark.parametrize('size', [100, 1000])
def test_cast_direct_labels_to_minimum_type_naive(size):
    pytest.importorskip('numba')
    data = np.arange(size, dtype=np.uint32)
    dtype = _accelerated_cmap.minimum_dtype_for_labels(size)
    cmap = DirectLabelColormap(
        color_dict={
            None: np.array([1, 1, 1, 1]),
            **{
                k: np.array([*v, 1])
                for k, v in zip(
                    range(size - 2),
                    product(np.linspace(0, 1, num=256), repeat=3),
                    strict=False,
                )
            },
        },
    )
    cmap.color_dict[None] = np.array([255, 255, 255, 255])
    res1 = _accelerated_cmap.labels_raw_to_texture_direct(data, cmap)
    res2 = _accelerated_cmap._labels_raw_to_texture_direct_numpy(data, cmap)
    npt.assert_array_equal(res1, res2)
    assert res1.dtype == dtype
    assert res2.dtype == dtype


def test_direct_colormap_with_no_selection():
    # Create a DirectLabelColormap with a simple color_dict
    color_dict = {
        1: np.array([1, 0, 0, 1]),
        2: np.array([0, 1, 0, 1]),
        None: np.array([0, 0, 0, 0]),
    }
    cmap = DirectLabelColormap(color_dict=color_dict)

    # Map a single value
    mapped = cmap.map(1)
    npt.assert_array_equal(mapped, np.array([1, 0, 0, 1]))

    # Map multiple values
    mapped = cmap.map(np.array([1, 2]))
    npt.assert_array_equal(mapped, np.array([[1, 0, 0, 1], [0, 1, 0, 1]]))


def test_direct_colormap_with_selection():
    # Create a DirectLabelColormap with a simple color_dict and a selection
    color_dict = {
        1: np.array([1, 0, 0, 1]),
        2: np.array([0, 1, 0, 1]),
        None: np.array([0, 0, 0, 0]),
    }
    cmap = DirectLabelColormap(
        color_dict=color_dict, use_selection=True, selection=1
    )

    # Map a single value
    mapped = cmap.map(1)
    npt.assert_array_equal(mapped, np.array([1, 0, 0, 1]))

    # Map a value that is not the selection
    mapped = cmap.map(2)
    npt.assert_array_equal(mapped, np.array([0, 0, 0, 0]))


def test_direct_colormap_with_invalid_values():
    # Create a DirectLabelColormap with a simple color_dict
    color_dict = {
        1: np.array([1, 0, 0, 1]),
        2: np.array([0, 1, 0, 1]),
        None: np.array([0, 0, 0, 0]),
    }
    cmap = DirectLabelColormap(color_dict=color_dict)

    # Map a value that is not in the color_dict
    mapped = cmap.map(3)
    npt.assert_array_equal(mapped, np.array([0, 0, 0, 0]))


def test_direct_colormap_with_values_outside_data_dtype():
    """https://github.com/napari/napari/pull/6998#issuecomment-2176070672"""
    color_dict = {
        1: np.array([1, 0, 1, 1]),
        2: np.array([0, 1, 0, 1]),
        257: np.array([1, 1, 1, 1]),
        None: np.array([0, 0, 0, 0]),
    }
    cmap = DirectLabelColormap(color_dict=color_dict)

    # Map an array with a dtype for which some dict values are out of range
    mapped = cmap.map(np.array([1], dtype=np.uint8))
    npt.assert_array_equal(mapped[0], color_dict[1].astype(mapped.dtype))


def test_direct_colormap_with_empty_color_dict():
    # Create a DirectLabelColormap with an empty color_dict
    with pytest.warns(Warning, match='color_dict did not provide'):
        DirectLabelColormap(color_dict={})


def test_direct_colormap_with_non_integer_values():
    # Create a DirectLabelColormap with a simple color_dict
    color_dict = {
        1: np.array([1, 0, 0, 1]),
        2: np.array([0, 1, 0, 1]),
        None: np.array([0, 0, 0, 0]),
    }
    cmap = DirectLabelColormap(color_dict=color_dict)

    # Map a float value
    with pytest.raises(TypeError, match='DirectLabelColormap can only'):
        cmap.map(1.5)

    # Map a string value
    with pytest.raises(TypeError, match='DirectLabelColormap can only'):
        cmap.map('1')


def test_direct_colormap_with_collision():
    # this test assumes that the the selected prime number for hash map size is 11
    color_dict = {
        1: np.array([1, 0, 0, 1]),
        12: np.array([0, 1, 0, 1]),
        23: np.array([0, 0, 1, 1]),
        None: np.array([0, 0, 0, 0]),
    }
    cmap = DirectLabelColormap(color_dict=color_dict)

    npt.assert_array_equal(cmap.map(1), np.array([1, 0, 0, 1]))
    npt.assert_array_equal(cmap.map(12), np.array([0, 1, 0, 1]))
    npt.assert_array_equal(cmap.map(23), np.array([0, 0, 1, 1]))


def test_direct_colormap_negative_values():
    # Create a DirectLabelColormap with a simple color_dict
    color_dict = {
        -1: np.array([1, 0, 0, 1]),
        -2: np.array([0, 1, 0, 1]),
        None: np.array([0, 0, 0, 0]),
    }
    cmap = DirectLabelColormap(color_dict=color_dict)

    # Map a single value
    mapped = cmap.map(np.int8(-1))
    npt.assert_array_equal(mapped, np.array([1, 0, 0, 1]))

    # Map multiple values
    mapped = cmap.map(np.array([-1, -2], dtype=np.int8))
    npt.assert_array_equal(mapped, np.array([[1, 0, 0, 1], [0, 1, 0, 1]]))


def test_direct_colormap_negative_values_numpy():
    color_dict = {
        -1: np.array([1, 0, 0, 1]),
        -2: np.array([0, 1, 0, 1]),
        None: np.array([0, 0, 0, 1]),
    }
    cmap = DirectLabelColormap(color_dict=color_dict)

    res = _accelerated_cmap._labels_raw_to_texture_direct_numpy(
        np.array([-1, -2, 5], dtype=np.int8), cmap
    )
    npt.assert_array_equal(res, [1, 2, 0])

    cmap.selection = -2
    cmap.use_selection = True

    res = _accelerated_cmap._labels_raw_to_texture_direct_numpy(
        np.array([-1, -2, 5], dtype=np.int8), cmap
    )
    npt.assert_array_equal(res, [0, 1, 0])


@pytest.mark.parametrize(
    'colormap_like',
    [
        ['red', 'blue'],
        [[1, 0, 0, 1], [0, 0, 1, 1]],
        {None: 'transparent', 1: 'red', 2: 'blue'},
        {None: [0, 0, 0, 0], 1: [1, 0, 0, 1], 2: [0, 0, 1, 1]},
        defaultdict(lambda: 'transparent', {1: 'red', 2: 'blue'}),
        CyclicLabelColormap(['red', 'blue']),
        DirectLabelColormap(
            color_dict={None: 'transparent', 1: 'red', 2: 'blue'}
        ),
        5,  # test ValueError
    ],
)
def test_normalize_label_colormap(colormap_like):
    if not isinstance(colormap_like, int):
        assert isinstance(
            _normalize_label_colormap(colormap_like), LabelColormapBase
        )
    else:
        with pytest.raises(ValueError, match='Unable to interpret'):
            _normalize_label_colormap(colormap_like)



from napari._vispy.overlays.base import (
    LayerOverlayMixin,
    ViewerOverlayMixin,
    VispyCanvasOverlay,
)
from napari._vispy.visuals.text import Text
from napari.components._viewer_constants import CanvasPosition
from napari.components.overlays import TextOverlay
from napari.settings import get_settings


class _VispyBaseTextOverlay(VispyCanvasOverlay):
    """Base class for vispy text overlays."""

    overlay: TextOverlay

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.node.font_size = self.overlay.font_size
        self.node.anchors = ('left', 'bottom')

        self.overlay.events.color.connect(self._on_color_change)
        self.overlay.events.box.connect(self._on_color_change)
        self.overlay.events.box_color.connect(self._on_color_change)
        self.overlay.events.font_size.connect(self._on_font_size_change)

        get_settings().appearance.events.theme.connect(self._on_color_change)
        self.viewer.events.theme.connect(self._on_color_change)

    def _connect_events(self):
        pass

    def _on_text_change(self):
        self.node.text = self.overlay.text
        self._on_position_change()

    def _on_visible_change(self):
        # ensure that dpi is updated when the scale bar is visible
        # this does not need to run _on_position_change because visibility
        # is already connected to the canvas callback by the canvas itself
        self._on_text_change()
        return super()._on_visible_change()

    def _on_color_change(self):
        self.node.color = (
            self.overlay.color
            if self.overlay.color is not None
            else self._get_fgcolor()
        )

    def _on_font_size_change(self):
        self.node.font_size = self.overlay.font_size
        self._on_position_change()

    def _on_position_change(self, event=None):
        position = self.overlay.position
        anchors = ('left', 'bottom')
        if position == CanvasPosition.TOP_LEFT:
            anchors = ('left', 'bottom')
        elif position == CanvasPosition.TOP_RIGHT:
            anchors = ('right', 'bottom')
        elif position == CanvasPosition.TOP_CENTER:
            anchors = ('center', 'bottom')
        elif position == CanvasPosition.BOTTOM_RIGHT:
            anchors = ('right', 'top')
        elif position == CanvasPosition.BOTTOM_LEFT:
            anchors = ('left', 'top')
        elif position == CanvasPosition.BOTTOM_CENTER:
            anchors = ('center', 'top')

        self.node.anchors = anchors
        self.node.font_size = self.overlay.font_size

        self.x_size, self.y_size = self.node.get_width_height()

        # depending on the canvas position, we need to change the position of the anchor itself
        # to ensure the text aligns properly e.g. left when on the left, and right when on the right
        x = y = 0.0
        if anchors[0] == 'right':
            x = self.x_size
        elif anchors[0] == 'center':
            x = self.x_size / 2

        if anchors[1] == 'top':
            y = self.y_size

        self.node.pos = (x, y)

        super()._on_position_change()

    def reset(self):
        super().reset()
        self._on_text_change()
        self._on_color_change()
        self._on_font_size_change()


class _VispyViewerTextOverlay(ViewerOverlayMixin, _VispyBaseTextOverlay):
    def __init__(self, **kwargs):
        font_manager = kwargs.get('font_manager')
        font_family = kwargs.get('font_family', 'OpenSans')
        super().__init__(
            node=Text(pos=(0, 0), font_manager=font_manager, face=font_family),
            **kwargs,
        )

        self._connect_events()
        self.reset()


class _VispyLayerTextOverlay(LayerOverlayMixin, _VispyBaseTextOverlay):
    def __init__(self, **kwargs):
        font_manager = kwargs.get('font_manager')
        font_family = kwargs.get('font_family', 'OpenSans')
        super().__init__(
            node=Text(pos=(0, 0), font_manager=font_manager, face=font_family),
            **kwargs,
        )

        self._connect_events()
        self.reset()


class VispyTextOverlay(_VispyViewerTextOverlay):
    def _connect_events(self):
        self.overlay.events.text.connect(self._on_text_change)

    def _on_text_change(self):
        self.node.text = self.overlay.text
        self._on_position_change()


class VispyLayerNameOverlay(_VispyLayerTextOverlay):
    def _connect_events(self):
        self.layer.events.name.connect(self._on_text_change)

    def _on_text_change(self):
        self.node.text = self.layer.name
        self._on_position_change()


class VispyCurrentSliceOverlay(_VispyViewerTextOverlay):
    def _connect_events(self):
        self.viewer.dims.events.connect(self._on_text_change)

    def _on_text_change(self):
        dims = self.viewer.dims
        lines = []
        for dim in dims.not_displayed:
            position = dims.point[dim]
            label = dims.axis_labels[dim]
            lines.append(f'{label} = {position:.5g}')

        self.node.text = '\n'.join(lines)
        self._on_position_change()



from __future__ import annotations

import contextlib
import json
import logging
import os
import warnings
from collections.abc import Mapping
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Literal
from warnings import warn

from pydantic import (
    AliasChoices,
    BaseModel,
    Field,
    PrivateAttr,
    TypeAdapter,
    ValidationError,
)
from pydantic.fields import FieldInfo
from pydantic_settings import (
    BaseSettings,
    EnvSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

from napari._pydantic_util import get_inner_type, get_origin
from napari.settings._yaml import PydanticYamlMixin
from napari.utils.compat import StrEnum
from napari.utils.events import EmitterGroup, EventedModel
from napari.utils.misc import StringEnum, deep_update
from napari.utils.translations import trans

_logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Union

    # TODO: needs to be fixed properly
    SettingsSourceCallable = Any

    from napari.utils.events import Event

    IntStr = Union[int, str]
    from pydantic.main import IncEx

    DictStrAny = dict[str, Any]
    MappingIntStrAny = Mapping[IntStr, Any]
    JSONable = str | list | dict | int | float | bool | None

Dict = dict  # rename, because EventedSettings has method dict


def _json_encode(
    dkt: dict[type, Callable[[Any], JSONable]],
) -> Callable[[Any], JSONable]:
    def json_encode(value: Any) -> JSONable:
        if type(value) in dkt:
            return dkt[type(value)](value)
        if isinstance(value, (StrEnum, StringEnum)):
            return value.value
        if isinstance(value, Path):
            return str(value)
        if isinstance(value, set):
            return [str(x) for x in value]

        raise TypeError(
            f'Object of type {type(value)} is not JSON serializable'
        )

    return json_encode


class EventedSettings(BaseSettings, EventedModel):
    """A variant of EventedModel designed for settings.

    Pydantic's BaseSettings model will attempt to determine the values of any
    fields not passed as keyword arguments by reading from the environment.
    """

    # provide config_path=None to prevent reading from disk.

    def __init__(self, **values: Any) -> None:
        super().__init__(**values)
        self.events.add(changed=None)
        self._connect(self)

    @staticmethod
    def _warn_restart(*_: Event) -> None:
        warn(
            trans._(
                'Restart required for this change to take effect.',
                deferred=True,
            )
        )

    def _connect(self, model: EventedModel, prefix: str = '') -> None:
        """Recursively connect and re-emit to all sub-fields."""
        for name, field in model.__class__.model_fields.items():
            attr = getattr(model, name)
            if isinstance(getattr(attr, 'events', None), EmitterGroup):
                path = f'{prefix}{name}'
                attr.events.connect(partial(self._on_sub_event, field=path))
                self._connect(attr, f'{path}.')

            extra = getattr(field, 'json_schema_extra', None)
            if extra is not None and extra.get('requires_restart', False):
                emitter = getattr(model.events, name)
                emitter.connect(self._warn_restart)

    def _on_sub_event(self, event: Event, field=None):
        """emit the field.attr name and new value"""
        if field:
            field += '.'
        value = getattr(event, 'value', None)
        self.events.changed(key=f'{field}{event._type}', value=value)


class _NotSetType:
    def __bool__(self) -> bool:
        return False


_NOT_SET = _NotSetType()


class FileConfigSettingsSource(PydanticBaseSettingsSource):
    """Class to load settings from a config file (yaml, json)."""

    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[Any, str, bool]:
        """Required to satisfy PydanticBaseSettingsSource interface.
        All logic is in __call__ method.
        """
        raise NotImplementedError

    def __call__(self) -> Dict[str, Any]:
        sources: list[str | Path] = list(
            getattr(self.settings_cls.model_config, 'sources', [])
        )
        default_cfg = self.settings_cls.model_fields['config_path'].default

        filepath = self.current_state.get('config_path')

        if filepath is not None and Path(filepath).exists():
            sources.append(filepath)

        if not sources:
            return {}

        data: dict = {}

        for path in sources:
            path_ = Path(path).expanduser().resolve()

            if not path_.is_file():
                # if it wasn't the `_config_path` stated in the BaseModel itself,
                # we warn, since this would have been user provided.
                if path_ != default_cfg:
                    _logger.warning(
                        trans._(
                            'Requested config path is not a file: {path}',
                            path=path_,
                        )
                    )
                continue
                # get loader for yaml/json
            if path_.suffix in {'.yaml', '.yml'}:
                load = __import__('yaml').safe_load
            elif path_.suffix == '.json':
                load = __import__('json').load
            else:
                warn(
                    trans._(
                        'Unrecognized file extension for config_path: {path}',
                        path=path,
                    )
                )
                continue

            try:
                # try to parse the config file into a dict
                new_data = load(path_.read_text()) or {}
            except Exception as err:  # noqa: BLE001
                _logger.warning(
                    trans._(
                        'The content of the napari settings file could not be read\n\nThe default settings will be used and the content of the file will be replaced the next time settings are changed.\n\nError:\n{err}',
                        deferred=True,
                        err=err,
                    )
                )
                continue
            assert isinstance(new_data, dict), path_.read_text()
            deep_update(data, new_data, copy=False)

        self.validate_settings_kwargs(data, filepath)

        data['config_file_settings'] = data.copy()
        return data

    def validate_settings_kwargs(
        self, data: dict[str, Any], path_: Path
    ) -> dict[str, Any]:
        try:
            # validate the data, passing config_path=None so we dont recurse
            # back to this point again.
            self.settings_cls(config_path=None, **data)
        except ValidationError as err:
            if self.settings_cls.model_config.get(
                'strict_config_check', False
            ):
                raise

            # if errors occur, we still want to boot, so we just remove bad keys
            errors = err.errors()
            msg = trans._(
                'Validation errors in config file(s).\nThe following fields have been reset to the default value:\n\n{errors}\n',
                deferred=True,
                errors=errors,  # TODO: is this good enough?
            )
            with contextlib.suppress(Exception):
                # we're about to nuke some settings, so just in case... try backup
                backup_path = path_.parent / f'{path_.stem}.BAK{path_.suffix}'
                backup_path.write_text(path_.read_text())

            _logger.warning(msg)
            try:
                _remove_bad_keys(data, [e.get('loc', ()) for e in errors])
            except KeyError:  # pragma: no cover
                _logger.warning(
                    trans._(
                        'Failed to remove validation errors from config file. Using defaults.'
                    )
                )
                data = {}
        return data


class NapariEnvSettingsSource(EnvSettingsSource):
    def __call__(self):
        res = super().__call__()
        env_lower = {k.lower(): v for k, v in os.environ.items()}
        self.scan_env_aliases(res, self.settings_cls, [], env_lower)

        res['env_settings'] = res.copy()
        return res

    def scan_env_aliases(
        self,
        dkt: dict[str, Any],
        class_: type[BaseSettings],
        path: list[str],
        env_dkt: dict[str, str],
    ):
        for field_name, field in class_.model_fields.items():
            if field.exclude:
                continue
            field_type = get_inner_type(field.annotation)
            if get_origin(field_type) is None and issubclass(
                field_type, BaseModel
            ):
                self.scan_env_aliases(
                    dkt, field_type, path + [field_name], env_dkt
                )
            if field.validation_alias is None:
                continue
            if (
                not isinstance(field.validation_alias, AliasChoices)
                or field.validation_alias.choices[0] != field_name
            ):
                raise ValueError(
                    f'Invalid validation alias for field {field_name} needs to be AliasChoices with first choice as field name'
                )
            for env_name in field.validation_alias.choices[1:]:
                if env_name not in env_dkt:
                    continue
                env_value = env_dkt[str(env_name)]
                value = TypeAdapter(field_type).validate_python(env_value)
                sub_dkt = dkt
                for sub in path:
                    sub_dkt = sub_dkt.setdefault(sub, {})
                if field_name in sub_dkt:
                    _logger.warning(
                        'Multiple environment variables found for %(field_name) at %(env_name) and %(existing_env_name). Using earlier value.',
                        extra={
                            'field_name': field_name,
                            'env_name': env_name,
                            'existing_env_name': field.validation_alias.choices,
                        },
                    )
                    continue
                sub_dkt[field_name] = value


class EventedConfigFileSettings(EventedSettings, PydanticYamlMixin):
    """This adds config read/write and yaml support to EventedSettings.

    If your settings class *only* needs to read variables from the environment,
    such as environment variables (but not a config file), then subclass from
    EventedSettings.
    """

    config_path: Path | _NotSetType | None = Field(default=None, exclude=True)
    env_settings: Dict = Field(
        default_factory=dict, exclude=True, repr=False, frozen=True
    )
    _save_on_change: bool = PrivateAttr(True)
    # this dict stores the data that came specifically from the config file.
    # it's populated in `config_file_settings_source` and
    # used in `_remove_env_settings`
    config_file_settings: Dict = Field(
        default_factory=dict, exclude=True, repr=False, frozen=True
    )

    # provide config_path=None to prevent reading from disk.
    def __init__(self, config_path=_NOT_SET, **values: Any) -> None:
        cfg = (
            config_path
            if config_path is not _NOT_SET
            else self.__class__.model_fields['config_path'].get_default()
        )
        # this line is here for usage in the `customise_sources` hook.  It
        # will be overwritten in __init__ by BaseModel._init_private_attributes
        # so we set it again after __init__.
        # self._config_path = _cfg
        if 'env_settings' in values:
            raise ValueError('env_settings is a reserved field name')
        super().__init__(config_path=cfg, **values)

    def _maybe_save(self):
        if self._save_on_change and self.config_path:
            self.save()

    def _on_sub_event(self, event, field=None):
        super()._on_sub_event(event, field)
        self._maybe_save()

    def dict(
        self,
        *,
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_env: bool = False,
    ) -> DictStrAny:
        """Return dict representation of the model.

        May optionally specify which fields to include or exclude.

        .. deprecated:: 0.7.0
              `dict` will be removed in napari 0.8.0 it is replaced by
              `model_dump` following pydantic 1 to 2 changes.
        """
        warnings.warn(
            'method `dict` is deprecated in 0.7.0 and will be removed in napari 0.8.0, use `model_dump` instead.',
            category=FutureWarning,
        )
        return self.model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            exclude_env=exclude_env,
        )

    def model_dump(
        self,
        *,
        mode: Literal['json', 'python'] | str = 'python',  # noqa: PYI051
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal['none', 'warn', 'error'] = True,
        serialize_as_any: bool = False,
        exclude_env: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Return dict representation of the model.

        May optionally specify which fields to include or exclude.

        For `exclude_env` kwarg: if True, will exclude any settings
        provided by environment variables.

        For other kwargs, see ``pydantic.BaseModel.model_dump`` docs:
        https://docs.pydantic.dev/latest/api/base_model/#pydantic.BaseModel.model_dump
        """
        data = super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            serialize_as_any=serialize_as_any,
            **kwargs,
        )
        if exclude_env:
            self._remove_env_settings(data)
        return data

    def _save_dict(self, **dict_kwargs: Any) -> DictStrAny:
        """The minimal dict representation that will be persisted to disk.

        By default, this will exclude settings values that match the default
        value, and will exclude values that were provided by environment
        variables.  Empty dicts will also be removed.
        """
        dict_kwargs.setdefault('exclude_defaults', True)
        dict_kwargs.setdefault('exclude_env', True)
        data = self.model_dump(**dict_kwargs)
        _remove_empty_dicts(data)
        return data

    def save(
        self, path: str | Path | _NotSetType | None = None, **dict_kwargs
    ):
        """Save current settings to path.

        By default, this will exclude settings values that match the default
        value, and will exclude values that were provided by environment
        variables.  (see `_save_dict` method.)
        """
        path = path or self.config_path
        # use insinstance so mypy is happy
        if not path or isinstance(path, _NotSetType):
            raise ValueError(
                trans._(
                    'No path provided in config or save argument.',
                    deferred=True,
                )
            )

        path = Path(path).expanduser().resolve()
        path.parent.mkdir(exist_ok=True, parents=True)
        self._dump(str(path), self._save_dict(**dict_kwargs))

    def _dump(self, path: str, data: Dict) -> None:
        """Encode and dump `data` to `path` using a path-appropriate encoder."""
        if str(path).endswith(('.yaml', '.yml')):
            data_ = self._yaml_dump(data)
        elif str(path).endswith('.json'):
            data_ = json.dumps(
                data, default=_json_encode(self.model_config['json_encoders'])
            )
        else:
            raise NotImplementedError(
                trans._(
                    'Can only currently dump to `.json` or `.yaml`, not {path!r}',
                    deferred=True,
                    path=path,
                )
            )
        with open(path, 'w') as target:
            target.write(data_)

    def _remove_env_settings(self, data):
        """Remove key:values from `data` that match settings from env vars.

        This is handy when we want to persist settings to disk without
        including settings that were provided by environment variables (which
        are usually more temporary).
        """
        if self.env_settings:
            _restore_config_data(
                data, self.env_settings, self.config_file_settings
            )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            NapariEnvSettingsSource(settings_cls),
            FileConfigSettingsSource(settings_cls),
            file_secret_settings,
        )

    model_config = SettingsConfigDict(
        strict_config_check=False,
    )


def _remove_bad_keys(data: dict, keys: list[tuple[int | str, ...]]):
    """Remove list of keys (as string tuples) from dict (in place).

    Parameters
    ----------
    data : dict
        dict to modify (will be modified inplace)
    keys : List[Tuple[str, ...]]
        list of possibly nested keys

    Examples
    --------

    >>> data = {'a': 1, 'b' : {'c': 2, 'd': 3}, 'e': 4}
    >>> keys = [('b', 'd'), ('e',)]
    >>> _remove_bad_keys(data, keys)
    >>> data
    {'a': 1, 'b': {'c': 2}}

    """
    for key in keys:
        if not key:
            continue  # pragma: no cover
        d = data
        while True:
            base, *key = key  # type: ignore
            if not key:
                break
            # since no pydantic fields will be integers, integers usually
            # mean we're indexing into a typed list. So remove the base key
            if isinstance(key[0], int):
                break
            d = d[base]
        del d[base]


def _restore_config_data(dct: dict, delete: dict, defaults: dict) -> dict:
    """delete nested dict keys, restore from defaults."""
    for k, v in delete.items():
        # recurse
        if isinstance(v, dict):
            dflt = defaults.get(k, {})
            if not isinstance(dflt, dict):
                dflt = {}
            _restore_config_data(dct.setdefault(k, {}), v, dflt)
        # restore from defaults if present, or just delete the key
        elif k in defaults:
            dct[k] = defaults[k]
        elif k in dct:
            del dct[k]

    return dct


def _remove_empty_dicts(dct: dict, recurse=True) -> dict:
    """Remove all (nested) keys with empty dict values from `dct`"""
    for k, v in list(dct.items()):
        if isinstance(v, Mapping) and recurse:
            _remove_empty_dicts(dct[k])
        if v == {}:
            del dct[k]
    return dct



"""
napari command line viewer.
"""

import argparse
import contextlib
import logging
import sys
import warnings
from ast import literal_eval
from pathlib import Path
from textwrap import wrap
from typing import Any

from napari import Viewer
from napari.errors import ReaderPluginError
from napari.utils._startup_script import _run_configured_startup_script
from napari.utils.misc import maybe_patch_conda_exe
from napari.utils.translations import trans


class InfoAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        # prevent unrelated INFO logs when doing "napari --info"

        from napari.utils import sys_info

        logging.basicConfig(level=logging.WARNING)
        print(sys_info())  # noqa: T201
        sys.exit()


class PluginInfoAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        # prevent unrelated INFO logs when doing "napari --info"
        logging.basicConfig(level=logging.WARNING)
        from npe2 import cli

        cli.list_(
            fields='name,version,npe2,contributions',
            sort='name',
            format='table',
        )
        sys.exit()


class CitationAction(argparse.Action):
    def __call__(self, *args, **kwargs):
        # prevent unrelated INFO logs when doing "napari --citation"
        from napari.utils import citation_text

        logging.basicConfig(level=logging.WARNING)
        print(citation_text)  # noqa: T201
        sys.exit()


def validate_unknown_args(unknown: list[str]) -> dict[str, Any]:
    """Convert a list of strings into a dict of valid kwargs for add_* methods.

    Will exit program if any of the arguments are unrecognized, or are
    malformed.  Converts string to python type using literal_eval.

    Parameters
    ----------
    unknown : List[str]
        a list of strings gathered as "unknown" arguments in argparse.

    Returns
    -------
    kwargs : Dict[str, Any]
        {key: val} dict suitable for the viewer.add_* methods where ``val``
        is a ``literal_eval`` result, or string.
    """

    from napari.components.viewer_model import valid_add_kwargs

    out: dict[str, Any] = {}
    valid = set.union(*valid_add_kwargs().values())
    for i, raw_arg in enumerate(unknown):
        if not raw_arg.startswith('--'):
            continue
        arg = raw_arg.lstrip('-')

        key, *values = arg.split('=', maxsplit=1)
        key = key.replace('-', '_')
        if key not in valid:
            sys.exit(f'error: unrecognized argument: {raw_arg}')

        if values:
            value = values[0]
        else:
            if len(unknown) <= i + 1 or unknown[i + 1].startswith('--'):
                sys.exit(f'error: argument {raw_arg} expected one argument')
            value = unknown[i + 1]
        with contextlib.suppress(Exception):
            value = literal_eval(value)

        out[key] = value
    return out


def parse_sys_argv():
    """Parse command line arguments."""

    from napari import __version__, layers
    from napari.components.viewer_model import valid_add_kwargs

    kwarg_options = []
    for layer_type, keys in valid_add_kwargs().items():
        kwarg_options.append(f'  {layer_type.title()}:')
        keys = {k.replace('_', '-') for k in keys}
        lines = wrap(', '.join(sorted(keys)), break_on_hyphens=False)
        kwarg_options.extend([f'    {line}' for line in lines])

    parser = argparse.ArgumentParser(
        usage=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="optional layer-type-specific arguments (precede with '--'):\n"
        + '\n'.join(kwarg_options),
    )
    parser.add_argument('paths', nargs='*', help='path(s) to view.')
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='increase output verbosity',
    )
    parser.add_argument(
        '-w',
        '--with',
        dest='with_',
        nargs='+',
        action='append',
        default=[],
        metavar=('PLUGIN_NAME', 'WIDGET_NAME'),
        help=(
            'open napari with dock widget from specified plugin name.'
            '(If plugin provides multiple dock widgets, widget name must also '
            'be provided). Use __all__ to open all dock widgets of a '
            'specified plugin. Multiple widgets are opened in tabs.'
        ),
    )
    parser.add_argument(
        '--version',
        action='version',
        version=f'napari version {__version__}',
    )
    parser.add_argument(
        '--info',
        action=InfoAction,
        nargs=0,
        help='show system information and exit',
    )
    parser.add_argument(
        '--plugin-info',
        action=PluginInfoAction,
        nargs=0,
        help='show information about plugins and exit',
    )
    parser.add_argument(
        '--citation',
        action=CitationAction,
        nargs=0,
        help='show citation information and exit',
    )
    # Allow multiple --stack options to be provided.
    # Each stack option will result in its own stack
    parser.add_argument(
        '--stack',
        action='append',
        nargs='*',
        default=[],
        help='concatenate multiple input files into a single stack. Can be provided multiple times for multiple stacks.',
    )
    parser.add_argument(
        '--plugin',
        help='specify plugin name when opening a file',
    )
    parser.add_argument(
        '--layer-type',
        metavar='TYPE',
        choices=set(layers.NAMES),
        help=(
            'force file to be interpreted as a specific layer type. '
            f'one of {set(layers.NAMES)}'
        ),
    )
    parser.add_argument(
        '--reset',
        action='store_true',
        help='reset settings to default values.',
    )
    parser.add_argument(
        '--settings-path',
        type=Path,
        help='use specific path to store and load settings.',
    )

    args, unknown = parser.parse_known_args()
    # this is a hack to allow using "=" as a key=value separator while also
    # allowing nargs='*' on the "paths" argument...
    for idx, item in enumerate(reversed(args.paths)):
        if item.startswith('--'):
            unknown.append(args.paths.pop(len(args.paths) - idx - 1))
    kwargs = validate_unknown_args(unknown) if unknown else {}

    return args, kwargs


def _run() -> None:
    from napari import run
    from napari.settings import get_settings

    """Main program."""
    args, kwargs = parse_sys_argv()

    # parse -v flags and set the appropriate logging level
    levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    level = levels[min(2, args.verbose)]  # prevent index error
    logging.basicConfig(
        level=level,
        format='%(asctime)s : %(levelname)s : %(threadName)s : %(message)s',
        datefmt='%H:%M:%S',
    )

    if args.reset:
        if args.settings_path:
            settings = get_settings(path=args.settings_path)
        else:
            settings = get_settings()
        settings.reset()
        settings.save()
        sys.exit('Resetting settings to default values.\n')

    if args.plugin:
        # make sure plugin is only used when files are specified
        if not args.paths:
            sys.exit(
                "error: The '--plugin' argument is only valid "
                'when providing a file name'
            )
        # I *think* that Qt is looking in sys.argv for a flag `--plugins`,
        # which emits "WARNING: No such plugin for spec 'builtins'"
        # so remove --plugin from sys.argv to prevent that warning
        sys.argv.remove('--plugin')

    if args.with_:
        from napari.plugins import (
            _initialize_plugins,
            _npe2,
        )

        # if a plugin widget has been requested, this will fail immediately
        # if the requested plugin/widget is not available.
        _initialize_plugins()

        npe2_plugins = []
        for plugin in args.with_:
            pname, *wnames = plugin
            for name, (w_pname, wnames) in _npe2.widget_iterator():
                if name == 'dock' and pname == w_pname:
                    npe2_plugins.append(plugin)
                    if '__all__' in wnames:
                        wnames = wnames
                    break

            if wnames:
                for wname in wnames:
                    _npe2.get_widget_contribution(pname, wname)
            else:
                _npe2.get_widget_contribution(pname)

    # viewer _must_  be kept around.
    # it will be referenced by the global window only
    # once napari has finished starting
    # but in the meantime if the garbage collector runs;
    # in a way that is machine, os, time (and likely weather dependant).
    # it will collect it and hang napari at start time.
    # don't show viewer until we've processed all the args
    viewer = Viewer(show=False)
    _run_configured_startup_script()

    # For backwards compatibility
    # If the --stack option is provided without additional arguments
    # just set stack to True similar to the previous store_true action
    if args.stack and len(args.stack) == 1 and len(args.stack[0]) == 0:
        warnings.warn(
            trans._(
                "The usage of the --stack option as a boolean is deprecated. Please use '--stack file1 file2 .. fileN' instead. It is now also possible to specify multiple stacks of files to stack '--stack file1 file2 --stack file3 file4 file5 --stack ..'. This warning will become an error in version 0.5.0.",
            ),
            DeprecationWarning,
            stacklevel=3,
        )
        args.stack = True
    try:
        viewer._window._qt_viewer._qt_open(
            args.paths,
            stack=args.stack,
            plugin=args.plugin,
            layer_type=args.layer_type,
            **kwargs,
        )
    except ReaderPluginError:
        logging.getLogger('napari').exception(
            'Loading %s with %s failed with errors',
            args.paths,
            args.plugin,
        )

    if args.with_:
        for plugin in npe2_plugins:
            pname, *wnames = plugin
            if '__all__' in wnames:
                for name, (
                    _pname,
                    wnames_collection,
                ) in _npe2.widget_iterator():
                    if name == 'dock' and pname == _pname:
                        wnames = wnames_collection
                        break

            if wnames:
                first_dock_widget = viewer.window.add_plugin_dock_widget(
                    pname, wnames[0], tabify=True
                )[0]
                for wname in wnames[1:]:
                    viewer.window.add_plugin_dock_widget(
                        pname, wname, tabify=True
                    )
                first_dock_widget.show()
                first_dock_widget.raise_()
            else:
                viewer.window.add_plugin_dock_widget(pname, tabify=True)

    # only necessary in bundled app, but see #3596
    from napari.utils.misc import (
        install_certifi_opener,
        running_as_constructor_app,
    )

    if running_as_constructor_app():
        install_certifi_opener()
        maybe_patch_conda_exe()
    # now that we've processed all the args, show viewer
    viewer.show()
    run(gui_exceptions=True)


def _run_plugin_module(mod, plugin_name):
    """Register `mod` as a plugin, find/create viewer, and run napari."""
    from napari import Viewer, run

    # now, check if a viewer was created, and if not, create one.
    for obj in mod.values():
        if isinstance(obj, Viewer):
            _v = obj
            break
    else:
        _v = Viewer()

    try:
        _v.window._qt_window.parent()
    except RuntimeError:
        # this script had a napari.run() in it, and the viewer has already been
        # used and cleaned up... if we eventually have "reusable viewers", we
        # can continue here
        return

    run()


def main():
    _run()


if __name__ == '__main__':
    sys.exit(main())



from __future__ import annotations

import warnings
from collections.abc import Sequence
from logging import getLogger
from typing import TYPE_CHECKING, Any

from napari.layers import Layer
from napari.plugins import _npe2
from napari.types import LayerData, PathLike
from napari.utils.translations import trans

logger = getLogger(__name__)
if TYPE_CHECKING:
    from npe2.manifest.contributions import WriterContribution


def read_data_with_plugins(
    paths: Sequence[PathLike],
    plugin: str | None = None,
    stack: bool = False,
) -> tuple[list[LayerData] | None, str | None]:
    """Iterate reader hooks and return first non-None LayerData or None.

    This function returns as soon as the path has been read successfully,
    while raising any errors occurring during the reading process, and
    providing useful error messages.

    Parameters
    ----------
    paths : str, or list of string
        The of path (file, directory, url) to open
    plugin : str, optional
        Name of a plugin to use.  If provided, will force ``path`` to be read
        with the specified ``plugin``.  If the requested plugin cannot read
        ``path``, an exception will be raised.
    stack : bool
        See `Viewer.open`

    Returns
    -------
    LayerData : list of tuples, or None
        LayerData that can be passed to :func:`Viewer._add_layer_from_data()
        <napari.components.viewer_model.ViewerModel._add_layer_from_data>`.
        ``LayerData`` is a list tuples, where each tuple is one of
        ``(data,)``, ``(data, meta)``, or ``(data, meta, layer_type)`` .

        If no reader plugins were found (or they all failed), returns ``None``
    reader : str, optional
        The name of the reader plugin that was used, or None if no reader was found.
    """
    if plugin == 'builtins':
        warnings.warn(
            trans._(
                'The "builtins" plugin name is deprecated and will not work in a future version. Please use "napari" instead.',
                deferred=True,
            ),
        )
        plugin = 'napari'

    assert isinstance(paths, list)
    if not stack:
        assert len(paths) == 1

    res = _npe2.read(paths, plugin, stack=stack)
    if res is not None:
        ld_, reader = res
        return [] if _is_null_layer_sentinel(ld_) else list(ld_), reader

    return [], None


def save_layers(
    path: str,
    layers: list[Layer],
    *,
    plugin: str | None = None,
    _writer: WriterContribution | None = None,
) -> list[str]:
    """Write list of layers or individual layer to a path using writer plugins.

    If ``plugin`` is provided, will attempt to use a writer declared by that
    plugin, assuming it is compatible with the given layer(s) and ``path``.
    Otherwise, will use first compatible writer.

    Parameters
    ----------
    path : str
        A filepath, directory, or URL to write.
    layers : List[layers.Layer]
        Non-empty List of layers to be saved. Warns when the list
        of layers is empty.
    plugin : str, optional
        Name of the plugin to use for saving. If None then the first compatible
        writer will be used.

    Returns
    -------
    list of str
        File paths of any files that were written.
    """
    writer_name = ''
    if layers:
        written, writer_name = _write_layers_with_plugins(
            path, layers, plugin_name=plugin, _writer=_writer
        )
    else:
        warnings.warn(trans._('No layers to write.'))
        return []

    # If written is empty, something went wrong.
    # Generate a warning to tell the user what it was.
    if not written:
        if writer_name:
            warnings.warn(
                trans._(
                    "Plugin '{name}' tried to save layers but did not return any written paths.",
                    deferred=True,
                    name=writer_name,
                )
            )
        elif plugin:
            warnings.warn(
                trans._(
                    "Given plugin '{name}' is not a valid writer for {path}.",
                    deferred=True,
                    name=plugin,
                    path=path,
                )
            )
        else:
            warnings.warn(
                trans._(
                    'No data written! A plugin could not be found to write these {length} layers to {path}.',
                    deferred=True,
                    length=len(layers),
                    path=path,
                )
            )

    return written


def _is_null_layer_sentinel(layer_data: Any) -> bool:
    """Checks if the layer data returned from a reader function indicates an
    empty file. The sentinel value used for this is ``[(None,)]``.

    Parameters
    ----------
    layer_data : LayerData
        The layer data returned from a reader function to check

    Returns
    -------
    bool
        True, if the layer_data indicates an empty file, False otherwise
    """
    return (
        isinstance(layer_data, list)
        and len(layer_data) == 1
        and isinstance(layer_data[0], tuple)
        and len(layer_data[0]) == 1
        and layer_data[0][0] is None
    )


def _write_layers_with_plugins(
    path: str,
    layers: list[Layer],
    *,
    plugin_name: str | None = None,
    _writer: WriterContribution | None = None,
) -> tuple[list[str], str]:
    """Write data from multiple layers data with a plugin.

    Calls out to _npe2.write_layers which will get the first
    writer compatible with the given layer combination and specified file
    extension in ``path``, and attempt to write the file(s).

    Parameters
    ----------
    path : str
        The path (file, directory, url) to write.
    layers : List of napari.layers.Layer
        List of napari layers to write.
    plugin_name : str, optional
        If provided, force the plugin manager to use this plugin's writer.
        If none is available, or if it is
        incapable of handling the layers, this function will fail.

    Returns
    -------
    (written paths, writer name) as Tuple[List[str],str]

    written paths: List[str]
        Empty list when no plugin was found, otherwise a list of file paths,
        if any, that were written.
    writer name: str
        Name of the plugin selected to write the data.
    """
    written_paths, writer_name = _npe2.write_layers(
        path, layers, plugin_name, _writer
    )
    if written_paths or writer_name:
        return (written_paths, writer_name)
    return ([], '')



from collections.abc import Collection
from types import NoneType, UnionType
from typing import (  # noqa
    Annotated,
    Any,
    Dict,
    List,
    Union,
    get_args,
    get_origin,
)

from pydantic import ConfigDict


class NapariConfigDict(ConfigDict):
    """Custom ConfigDict for napari models.

    add dependencies attribute to track field dependencies
    """

    # FIXME: make it part of public api?
    dependencies: dict[str, Collection[str]] | None


def iter_inner_types(type_: Any) -> Any:
    origin = get_origin(type_)
    args = get_args(type_)
    if origin in (list, List):  # noqa
        yield from iter_inner_types(args[0])
    elif origin in (dict, Dict):  # noqa
        yield from iter_inner_types(args[1])
    elif origin is Annotated:
        yield from iter_inner_types(args[0])
    elif origin in (UnionType, Union):
        for arg in args:
            yield from iter_inner_types(arg)
    elif type_ is not NoneType:
        yield type_


def get_inner_type(type_: Any) -> Any:
    """Roughly replacing pydantic.v1 Field.type_"""
    return Union[tuple(iter_inner_types(type_))]


def get_outer_type(type_: Any) -> Any:
    """Roughly replacing pydantic.v1 Field.outer_type_"""
    origin = get_origin(type_)
    args = get_args(type_)
    if origin in (UnionType, Union):
        # filter args to remove optional None
        args = tuple(filter(lambda t: t is not NoneType, get_args(type_)))
        if len(args) == 1:
            # it was just optional, pretend there was no None
            return get_outer_type(args[0])
        # It's an actual union of types, so there's no "outer type"
        return None
    if origin is not None:
        return origin
    return type_


def is_list_type(type_: Any) -> bool:
    """Roughly replacing pydantic.v1 comparison to SHAPE_LIST"""
    return get_outer_type(type_) is list


__all__ = (
    'get_inner_type',
    'get_outer_type',
    'is_list_type',
    'iter_inner_types',
)
